"""
This module defines Executor classes.

Executors are used to submit and monitor jobs.
"""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import glob
import os
import subprocess

from themis.utils import (
    URI_ENV_VAR,
    RUNID_ENV_VAR,
    URI_SPLITCHAR,
    SETUPDIR_ENV_VAR,
    EXECDIR_ENV_VAR,
)

try:
    import concurrent.futures as cf
    import flux
    import flux.job
except (ImportError, SyntaxError):
    FLUX_BINDINGS_AVAIL = False
else:
    FLUX_BINDINGS_AVAIL = True


RUN_LOG_FILE = 'run' + os.extsep + 'log'
LOG_EXT = os.extsep + 'log'
RUNLOG_GLBR = 'run' + '*' + LOG_EXT

def _exitstatus_to_returncode(status):
    """Convert an `os.wait` exit status to a returncode."""
    if os.WIFEXITED(status):
        return os.WEXITSTATUS(status)
    if os.WIFSIGNALED(status):
        return -os.WTERMSIG(status)
    return -256


class Executor(object):
    """Abstract class that provides methods to submit and wait on jobs
    asynchronously."""

    def __init__(self, setup_dir, themis_uri):
        self.themis_uri = URI_SPLITCHAR.join(themis_uri)
        self.setup_dir = setup_dir
        self.exec_dir = os.getcwd()

    # pylint: disable=too-many-arguments
    def submit(self, run_id, step):
        """Submit a job and return a job ID.

        Stdout and stderr for the job will be redirected to `cwd/run.log`.

        :param step: The themis.utils.Step to launch.

        :returns: an integer identifying the job.
        """
        raise NotImplementedError()

    def wait(self, timeout=None):
        """Wait for any job to complete and return the pair (jobid, exit_status).

        The exit_status entry will be either a process exit code or an exception.

        If there are no jobs, return `None`.

        :param timeout: the number of seconds to wait for a job to complete.
            If None, wait indefinitely.
        """
        raise NotImplementedError()

    def kill(self, jobid):
        """Cancel or kill a job."""


class ProcessExecutor(Executor):
    """Submit jobs as individual processes."""

    def __init__(self, setup_dir, themis_uri, command_builder):
        super(ProcessExecutor, self).__init__(setup_dir, themis_uri)
        self._command_builder = command_builder
        self._processes = {}
        self._environ = dict(os.environ)
        self._environ[URI_ENV_VAR] = self.themis_uri
        self._environ[SETUPDIR_ENV_VAR] = self.setup_dir


    # pylint: disable=too-many-arguments
    def submit(self, run_id, step):
        """Submit an application as a process. Return the process ID (pid)."""
        command = self._command_builder(step)  # get the prefix
        command.extend(step.args)
        env = dict(self._environ)
        env[RUNID_ENV_VAR] = str(run_id)

        print("source/executors.py submit")
        print(step)
        print(step.cwd)
        cwd = step.cwd[:-1]

        if os.path.isfile(os.path.join(cwd, RUN_LOG_FILE)):
            count = len(glob.glob(os.path.join(cwd, RUNLOG_GLBR)))
            os.rename(os.path.join(cwd,  RUN_LOG_FILE),
                     "%s_%04d%s" % (os.path.join(cwd,  'run'), count, LOG_EXT))

        with open(os.path.join(step.cwd, RUN_LOG_FILE), "a") as run_log:
            process = subprocess.Popen(
                command,
                cwd=step.cwd,
                stdout=run_log,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                env=env,
            )
        self._processes[process.pid] = process
        return process.pid

    def wait(self, timeout=None):
        """Wait for a job to a complete and return its returncode.

        Any timeout value other than `None` has no effect.
        """
        if not self._processes:
            return None
        if timeout is None and hasattr(os, "wait"):
            pid, exit_status = os.wait()
            completed_proc = self._processes.pop(pid, None)
            if completed_proc is not None:
                completed_proc.poll()
                return (pid, _exitstatus_to_returncode(exit_status))
        else:
            for proc in self._processes.values():
                if proc.poll() is not None:
                    completed_proc = proc
                    self._processes.pop(completed_proc.pid)
                    return (completed_proc.pid, completed_proc.returncode)
        return None

    def kill(self, jobid):
        """Kill or cancel a job."""
        self._processes[jobid].terminate()


class FluxBindingsExecutor(Executor):
    """Submit applications via `flux.job.FluxExecutor`."""

    def __init__(self, setup_dir, themis_uri):
        super(FluxBindingsExecutor, self).__init__(setup_dir, themis_uri)
        self._executor = flux.job.FluxExecutor()
        self._futures = {}
        self._next_jobid = 0
        self._flux_handle = flux.Flux()  # for cancelling futures
        self._environ = dict(os.environ)
        self._environ[URI_ENV_VAR] = self.themis_uri
        self._environ[SETUPDIR_ENV_VAR] = self.setup_dir
        self._environ[EXECDIR_ENV_VAR] = self.exec_dir

    # pylint: disable=too-many-arguments
    def submit(self, run_id, step):
        """Submit an application via Flux. Return an integer."""
        if step.batch_script:
            jobspec = flux.job.JobspecV1.from_nest_command(
                step.args,
                num_slots=step.tasks,
                cores_per_slot=step.cores_per_task,
                gpus_per_slot=step.gpus_per_task,
            )
        else:
            jobspec = flux.job.JobspecV1.from_command(
                step.args,
                num_tasks=step.tasks,
                cores_per_task=step.cores_per_task,
                gpus_per_task=step.gpus_per_task,
            )
        env = dict(self._environ)
        env[RUNID_ENV_VAR] = str(run_id)
        jobspec.environment = env
        if step.cwd is not None:
            jobspec.cwd = step.cwd
        if step.timeout is not None and step.timeout > 0:
            jobspec.duration = float(step.timeout * 60)

        cwd = step.cwd
        if os.path.isfile(os.path.join(cwd, RUN_LOG_FILE)):
            count = len(glob.glob(os.path.join(cwd, RUNLOG_GLBR)))
            os.rename(os.path.join(cwd,  RUN_LOG_FILE),
                     "%s_%04d%s" % (os.path.join(cwd,  'run'), count, LOG_EXT))

        jobspec.stdout = os.path.join(step.cwd, RUN_LOG_FILE)
        self._futures[self._executor.submit(jobspec)] = self._next_jobid
        self._next_jobid += 1
        return self._next_jobid - 1

    def wait(self, timeout=None):
        """Wait for a job to complete."""
        if not self._futures:
            return None
        done, _ = cf.wait(self._futures, timeout, return_when=cf.FIRST_COMPLETED)
        if done:
            completed_fut = done.pop()
            completed_jobid = self._futures.pop(completed_fut)
            try:
                result = completed_fut.result()
            except Exception as exc:  # pylint: disable=broad-except
                result = exc
            return (completed_jobid, result)
        return None

    def kill(self, jobid):
        """Kill/cancel a job."""
        for fut, iter_jobid in self._futures.items():
            if jobid == iter_jobid:
                if not fut.cancel():
                    flux.job.cancel(self._flux_handle, fut.jobid())
                return



"""Networking classes for communicating between workers and their sub-scripts."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import collections
import threading

from themis.backend import network
from themis.versions import pickle, UserDict


class ThreadSafeDict(UserDict, object):  # pylint: disable=too-many-ancestors
    """Dictionary that protects certain operations with an RLock object."""

    def __init__(self, *args, **kwargs):
        self.lock = threading.RLock()
        super(ThreadSafeDict, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        with self.lock:
            return super(ThreadSafeDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        with self.lock:
            return super(ThreadSafeDict, self).__getitem__(key)

    def __delitem__(self, key):
        with self.lock:
            return super(ThreadSafeDict, self).__delitem__(key)

    def items(self):
        with self.lock:
            return tuple(super(ThreadSafeDict, self).items())

    def keys(self):
        with self.lock:
            return tuple(super(ThreadSafeDict, self).keys())


class WorkerServerRequestHandler(network.BaseRequestHandler):
    """Handles requests to the server. Each request instantiates an instance."""

    def handle_results(self, request, content):
        """Handle a worker's results POST request"""
        self.server.themis_results.append(
            (int(request["runid"][0]), pickle.loads(content))
        )

    def handle_setup(self, request, content):  # pylint: disable=unused-argument
        """Handle a worker's setup GET request"""
        return self.server.themis_run_infos[int(request["runid"][0])]


class WorkerServer(network.BaseServer):
    """Serve database requests needed by a workers' subscripts."""

    def __init__(self, cert_path, app_spec):
        super(WorkerServer, self).__init__(WorkerServerRequestHandler, cert_path)
        self.app_spec = app_spec
        self.results = collections.deque()
        self.run_infos = ThreadSafeDict()
        self.server.themis_run_infos = self.run_infos
        self.server.themis_results = self.results

    def submit(self, run_info_iterable):
        """Add the RunInfo objects to the cache for distribution."""
        for run_info in run_info_iterable:
            self.run_infos[run_info.run_id] = pickle.dumps(
                {"run_info": run_info, "app_spec": self.app_spec},
                protocol=pickle.HIGHEST_PROTOCOL,
            )

    def remove_completed_runs(self, run_ids_to_delete):
        """Remove completed runs from the cache.

        Allows the server to free up memory; otherwise the server's memory
        would continue growing the longer it runs.
        """
        for run_id in run_ids_to_delete:
            del self.run_infos[run_id]

    def save_progress(self, ensemble_db):
        """Write the server's cache of results to the database.

        Since the results are not used anywhere else, the
        results can be deleted as soon as they are written.
        """
        results_as_list = []
        while self.results:
            results_as_list.append(self.results.popleft())
        if results_as_list:
            ensemble_db.add_result(
                tuple(pair[0] for pair in results_as_list),
                tuple(pair[1] for pair in results_as_list),
            )


class WorkerClient(network.BaseClient):
    """Class for making calls to a WorkerServer instance."""

    def get_worker_setup(self, run_id):
        """Get the setup for a run from the server."""
        payload = self._get_pickled_data("setup", {"runid": run_id})
        return (payload["app_spec"], payload["run_info"])

    def add_result(self, run_id, result):
        """Send the result to the server."""
        self._send_data_pickle("results", {"runid": run_id}, result)



"""
This packagew defines the EnsembleDatabase class and supporting functions.

The EnsembleDatabase class is used to connect and interact with
the ensemble's backend. This backend is used all over this package--
by both workers and by the user-facing API.

The backend (whatever it is) is used both as a simple store of information
and as concurrency control. Most of the operations `must` be atomic.

The reference EnsembleDatabase implementation is the SQLiteDatabase class,
defined in this module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import json

from themis.database.statusstore import SQLiteStatusStore
from themis.database import sqlitedb


APP_SPEC_NAME = "ensemble_spec.json"

STATUS_STORE_NAME = "ensemble_status.db"


def get_status_store(where=os.curdir):
    """Return the StatusStore instance to be used for the ensemble."""
    return SQLiteStatusStore(os.path.abspath(os.path.join(where, STATUS_STORE_NAME)))


def default_database(app_spec=None):
    """Return the EnsembleDatabase instance to be used for the ensemble.

    :param app_spec: the app_spec dictionary, which holds information
        about the type and location of the database. If None,
        it will be collected via the get_app_spec function.
    """
    if app_spec is None:
        app_spec = get_app_spec()
    if app_spec["database"]["type"] == "sqlite":
        return sqlitedb.SQLiteDatabase(app_spec["database"]["path"])
    raise ValueError("Database type not recognized: {}".format(app_spec["database"]))


def write_app_spec(app_spec, where=os.curdir):
    """Pickle the app_spec dictionary and dump to a file"""
    with open(os.path.join(where, APP_SPEC_NAME), "w") as file_handle:
        json.dump(app_spec, file_handle, indent=4)


def get_app_spec(where=os.curdir):
    """Collect the app_spec dictionary from the file system.

    :param where: the directory to seach for the json'd app_spec
        dictionary. If None, search in the cwd.
    """
    with open(os.path.join(where, APP_SPEC_NAME), "r") as file_handle:
        return json.load(file_handle)



"""
This package contains the Themis ensemble manager.

Themis is used to run an HPC application in parallel
inside of a single allocation. Usually this goes along with
varying the inputs of each run of the application.

The primary user interface to this package is contained in the
``themis`` module, which module provides the `Themis` class
for creating, restarting, and monitoring ensembles.

The ``themis.user_utils`` module provides an interface for passing
information back and forth between the application interface and Themis.
This module is only meant to be used by application interface
functions, which are themselves called by the ensemble manager while
it runs. Consult the documentation for a complete explanation.

The ``themis.laf`` module provides a Python and command-line interface to
launching allow_multiple batch scripts to a resource manager. Run that module directly
to produce the command-line interface.

Any other modules contained in this package are not intended for user use.

Running this package directly produces the command-line interface for
the ensemble manager.

Consult the documentation on each of those classes for information on usage.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import glob
import time
import json
import csv
import numbers
import shutil

from themis import database
from themis.database.utils import EnsembleDatabase
from themis import utils
from themis.utils import Run, CompositeRun, Step
from themis.resource import allocators
from themis import resource
from themis import versions
from themis import backend
from themis.backend.worker import prepper, finisher


__all__ = ["Themis", "user_utils", "runtime", "laf"]

__version__ = 1.2

_DEFAULT_MAX_CONCURRENCY = 100


def _get_app_spec_or_raise(setup_dir):
    try:
        return database.get_app_spec(setup_dir)
    except (IOError, OSError):
        raise ValueError("setup not found in {!r}".format(setup_dir))


def _validate_app_interface(app_interface):
    """Attempt to import the app interface"""
    try:
        utils.import_app_interface(app_interface)
    except ImportError as err:
        raise ImportError("Couldn't import the application interface: " + str(err))


class Themis(object):  # pylint: disable=too-many-public-methods
    """Represents and interfaces with an existing Themis ensemble.

    When creating a new ensemble, use one of the ``Themis.create*`` class methods,
    which will return a ``themis.Themis`` instance.
    The class constructor is used for getting a handle on an existing Themis
    ensemble---i.e. after ``create`` has been called.

    Use one of the ``Themis.execute_*`` methods to start an ensemble. The
    ``dryrun`` method is useful for checking that everything is in order.
    """

    #: This status matches runs which have not yet completed.
    RUN_QUEUED = EnsembleDatabase.RUN_QUEUED
    #: This status matches runs which have completed successfully.
    RUN_SUCCESS = EnsembleDatabase.RUN_SUCCESS
    #: This status matches runs which have failed.
    RUN_FAILURE = EnsembleDatabase.RUN_FAILURE
    #: This status matches runs which have been killed manually.
    RUN_KILLED = EnsembleDatabase.RUN_KILLED
    #: This status matches runs which returned one of the ``abort_on``.
    RUN_ABORTED = EnsembleDatabase.RUN_ABORTED

    STATUS_ENUMS = (RUN_QUEUED, RUN_SUCCESS, RUN_FAILURE, RUN_KILLED, RUN_ABORTED)

    # maps enums to human-readable strings, and vice versa
    _TRANSLATOR = {
        RUN_QUEUED: "queued",
        RUN_SUCCESS: "successful",
        RUN_FAILURE: "failed",
        RUN_KILLED: "killed",
        RUN_ABORTED: "aborted",
    }
    _TRANSLATOR.update({val: key for key, val in _TRANSLATOR.items()})

    @classmethod
    def translate_enum(cls, enum):
        """Convert a ``RUN_*`` enum to a human-readable string, or vice-versa."""
        return cls._TRANSLATOR[enum]

    @classmethod
    def create(  # pylint: disable=too-many-arguments,too-many-locals
        cls,
        application=None,
        runs=None,
        run_parse=None,
        run_copy=None,
        run_symlink=None,
        run_dir_names=None,
        app_interface=None,
        setup_dir=utils.DEFAULT_SETUP_DIR,
        max_restarts=0,
        max_failed_runs=None,
        app_is_batch_script=True,
        use_flux=False,
        abort_on=None,
    ):
        """Create a new Themis ensemble.

        Returns a ``themis.Themis`` instance representing the new ensemble.

        Only when one of the ``execute_*`` method is called does the ensemble actually
        begin.

        :param application: Path to the application to be executed. For instance,
            ``"/usr/bin/ls"``, ``"ls"``, or ``"../../bin/ls"`` might all be valid.
            This option is only relevant if ``themis.Run``
            (rather than ``CompositeRun``) objects are passed to this ensemble.
            If only ``CompositeRun`` objects are passed,
            the application is taken on a per-step basis from ``step.args[0]``.
        :type application: optional, str
        :param runs: An iterable of ``themis.CompositeRun`` objects,
            representing the runs of this ensemble.
            Runs within the ensemble will be executed concurrently.
            Each run is assigned a unique integer ID
            (the "run ID") corresponding to its index in the iterable. If no runs
            are passed, an empty ensemble is created.
        :type runs: optional
        :param run_parse: A file path or iterable of file paths. Unix-style path
            patterns are supported as well. The files specified should be text files;
            they will be hard-copied into the run directories and parsed.
            See :ref:`here <text_replacement>` for more. ``None``, the
            default, specifies that no files will be parsed.
        :type run_parse: str or iterable of str, optional
        :param run_copy: A file path or iterable of file paths. Unix-style path
            patterns are supported as well. The files/directories
            specified will be hard-copied into the run directories. ``None``, the
            default, specifies that no files will be copied.
        :type run_copy: str or iterable of str, optional
        :param run_symlink: A file path or iterable of file paths. Unix-style path
            patterns are supported as well. The files/directories
            specified will be symlinked into the run directories. ``None``, the
            default, specifies that no files will be symlinked.
        :type run_symlink: str or iterable of str, optional
        :param run_dir_names: The file system paths of the run directories.
            This argument should be a `python format string
            <https://docs.python.org/3/library/string.html#formatstrings>`_, where
            the field names correspond to the names of the variables in the samples
            argument. For instance, if the variables in the samples argument are
            "hydrostatics" and "viscocity", you might pass in the string
            ``"hydro={hydrostatics}/visc={viscocity}"``. This string will be
            formatted each run to yield the run directory; so one directory
            might be ``hydro=17.6/visc=35``. Note that
            the posix directory separator character "/"
            in the example string means that the resulting run directory
            will be in fact a sequence of two directories.
            The default value of ``None`` lets the naming
            scheme be determined internally.
        :type run_dir_names: str, optional
        :param app_interface: A file path to an
            :ref:`application interface <application_interface>` module.
        :type app_interface: str, optional
        :param setup_dir: A file system path indicating the directory in which to
            put the ensemble's setup files.
        :type setup_dir: str, optional
        :param max_restarts: the maximum number of times an individual run should be
            restarted if it fails---i.e. if the application exits with a
            nonzero return code.
            Setting this argument to ``None`` allows infinite restarts.
        :type max_restarts: int, optional
        :param max_failed_runs: the maximum total failures across the ensemble.
            If this number is exceeded, the ensemble will abort. The default
            of ``None`` allows infinite failures.
        :type max_failed_runs: int, optional
        :param app_is_batch_script: A boolean indicating whether the
            application given by the ``application`` parameter is
            a script that will launch parallel applications. For instance, if
            the application is a Slurm batch script that contains sruns,
            then this argument should be set to True. See the
            :ref:`batch script info page <batch_script_howto>` for more detail.
        :type app_is_batch_script: bool, optional
        :param use_flux: a boolean indicating whether to use Flux as the resource
            manager for the ensemble instead of the machine's
            native manager (e.g. Slurm).
            Can also be a string identifying the path to the Flux installation to use.
        :type use_flux: bool or str, optional
        :param abort_on: a sequence of positive integers. Each integer
            identifies a OS process return code. If ``application`` exits with
            one of those return codes, that run will be marked as ``RUN_ABORT``
            and will not be restarted.
        :returns: a ``Themis`` object representing a new ensemble
        :raises ValueError: if an ensemble exists in ``setup_dir``.
        """
        if cls.exists(setup_dir):
            raise Exception(
                "It looks like an ensemble was already "
                "created in the directory {0!r}.".format(setup_dir)
            )
        if not os.path.exists(setup_dir):
            os.makedirs(setup_dir)
        setup_dir = os.path.abspath(setup_dir)
        _validate_app_interface(app_interface)
        # initialize self
        if run_dir_names is None:
            run_dir_names = os.path.abspath(utils.DEFAULT_RUN_DIR_NAMES)
        else:
            run_dir_names = os.path.abspath(run_dir_names)
        if use_flux:
            resource_mgr = resource.identify_resource_manager(resource.Flux.identifier)
            flux_path = resource.validate_flux_path(use_flux)
        else:
            resource_mgr = resource.identify_resource_manager()
            flux_path = None
        run_parse, run_copy, run_symlink = utils.validate_run_files(
            run_parse, run_copy, run_symlink,
        )
        if app_is_batch_script and application:
            try:  # if the application is both parsed and symlinked, things get ugly
                run_symlink.remove(utils.validate_application(application))
            except ValueError:
                pass
        # many arguments are only needed on the compute nodes, and so are just stored
        # in the database; store them in a dict, to be later passed to the db
        spec = {"on_completion": None, "root_dir": os.getcwd(), "save_interval": 30}
        spec["application"] = (
            utils.validate_application(application) if application else None
        )
        spec["run_parse"] = run_parse
        spec["run_symlink"] = run_symlink
        spec["run_copy"] = run_copy
        spec["version_created"] = sys.version_info.major
        spec["app_interface"] = (
            None if app_interface is None else utils.existing_file(app_interface)
        )
        spec["resource_mgr"] = resource_mgr.identifier
        spec["flux_path"] = flux_path
        spec["setup_dir"] = setup_dir
        spec["run_dir_names"] = run_dir_names
        spec["max_restarts"] = utils.range_check(
            int(utils.convert_none(max_restarts, -1)), min_val=-1, name="max_restarts"
        )
        spec["max_failed_runs"] = utils.range_check(
            int(utils.convert_none(max_failed_runs, -1)),
            min_val=-1,
            name="max_failed_runs",
        )
        spec["app_is_batch_script"] = bool(app_is_batch_script)
        spec["database"] = {
            "type": "sqlite",
            "path": os.path.join(spec["setup_dir"], "ensemble_database.db"),
            "files": (os.path.join(spec["setup_dir"], "ensemble_database.db"),),
        }
        abort_on = () if abort_on is None else abort_on
        if not all((isinstance(code, int) and code > 0 for code in abort_on)):
            raise ValueError("`abort_on` must contain only positive integers")
        spec["abort_on"] = abort_on
        database.write_app_spec(spec, spec["setup_dir"])
        database.default_database(spec).create()
        database.get_status_store(spec["setup_dir"]).create()
        instance = cls(spec["setup_dir"])
        if runs:
            instance.add_runs(runs)
        return instance

    @classmethod
    def create_overwrite(cls, *args, **kwargs):
        """Create a new ensemble, removing an old one if it exists.

        Takes the same arguments as ``Themis.create()``.

        :returns: a ``Themis`` object representing a new ensemble
        """
        setup_dir = kwargs.get("setup_dir", utils.DEFAULT_SETUP_DIR)
        if cls.exists(setup_dir):
            cls.clear(setup_dir)
        return cls.create(*args, **kwargs)

    @classmethod
    def create_resume(cls, *args, **kwargs):
        """Create a new ensemble, or get a handle to an existing one.

        If an ensemble exists, return a ``Themis`` handle to it.
        Otherwise, create a new ensemble with the given arguments
        and return a ``Themis`` handle to it.

        Takes the same arguments as ``Themis.create()``.

        :returns: a ``Themis`` object representing either a new or an existing ensemble
        """
        setup_dir = kwargs.get("setup_dir", utils.DEFAULT_SETUP_DIR)
        if cls.exists(setup_dir):
            return cls(setup_dir)
        return cls.create(*args, **kwargs)

    @classmethod
    def clear(cls, setup_dir=utils.DEFAULT_SETUP_DIR):
        """Clear files generated by the ensemble manager for its own use.

        This function won't remove any user-provided files--for instance,
        input decks or required files. However, it will remove the
        ensemble manager's own log files, and any other files it uses
        to store its information. This includes stored samples and
        results; therefore, be careful that this is what you actually
        want to do.
        """
        if not cls.exists(setup_dir):
            return None
        app_spec = database.get_app_spec(setup_dir)
        expected_files = [
            os.path.join(setup_dir, database.APP_SPEC_NAME),
            os.path.join(setup_dir, database.STATUS_STORE_NAME),
        ]
        expected_files.extend(glob.glob(os.path.join(setup_dir, "themis_*.log")))
        expected_files.extend(
            glob.glob(os.path.join(setup_dir, "ensemble_launch_script.sh"))
        )
        expected_files.extend(app_spec["database"]["files"])
        for path in expected_files:
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)
        return None

    @classmethod
    def exists(cls, setup_dir=utils.DEFAULT_SETUP_DIR):
        """Return ``True`` if an ensemble's setup exists in ``setup_dir``.

        :param setup_dir: has the same meaning as the ``setup_dir`` parameter
            to the ``themis.Themis.__init__`` method.
        """
        try:
            cls(setup_dir)
        except ValueError:
            return False
        else:
            return True

    def __init__(self, setup_dir=utils.DEFAULT_SETUP_DIR):
        """Constructor. Used to get a handle to an existing Themis ensemble."""
        self._setup_dir = os.path.abspath(setup_dir)
        self._spec = _get_app_spec_or_raise(self._setup_dir)
        self._status_store = database.get_status_store(self._setup_dir)
        self._db = database.default_database(self._spec)
        if self._spec["version_created"] > sys.version_info.major:
            raise RuntimeError(
                "Ensemble created with Python {} but running in Python {}".format(
                    self._spec["version_created"], sys.version_info.major
                )
            )

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}({!r})".format(type(self).__name__, self._spec["setup_dir"])

    @property
    def setup_dir(self):
        """Return the setup directory where Themis stores its data."""
        return self._spec["setup_dir"]

    def execute_local(  # pylint: disable=too-many-arguments
        self,
        blocking=False,
        timeout=None,
        parallelism=None,
        allow_multiple=False,
        max_concurrency=None,
    ):
        """Execute the ensemble locally. Return ``None``.

        :param blocking: if ``True``, block until the ensemble exits.
        :type blocking: bool
        :param timeout: length in minutes to run Themis for. Default is no time limit
        :type timeout: int, optional
        :param parallelism: if set to a positive integer N, reserve N cores for Themis
            itself, and run Themis in parallel on those cores. If you suspect Themis's
            performance is bottlenecking your ensemble, increase the parallelism.
        :type parallelism: int, optional
        :param allow_multiple: if ``True``, enable multiple concurrent executions
            (i.e. multiple simultaneous calls to ``execute_alloc`` or
            ``execute_local``). As this makes certain checks and behavior
            impossible, it is not the default.
        :type allow_multiple: bool, optional
        :param max_concurrency: the maximum number of concurrent runs Themis should
            allow at any given time. Too low a number may cause your ensemble to
            progress very slowly, much too large a number may eventually cause
            performance degradation on certain systems.
            If ``None``, pick a reasonable default value.
        :type max_concurrency: int, optional
        """
        self._execute(
            allocators.Allocation(timeout=timeout, repeats=0),
            allocator=allocators.InteractiveAllocator(),
            blocking=blocking,
            parallelism=parallelism,
            allow_multiple=allow_multiple,
            max_concurrency=max_concurrency,
        )

    def execute_alloc(  # pylint: disable=too-many-arguments
        self,
        nodes=1,
        partition=None,
        bank=None,
        name="themis",
        timeout=None,
        repeats=0,
        parallelism=None,
        early_stop=None,
        allow_multiple=False,
        max_concurrency=None,
    ):
        """Request an allocation and launch the ensemble within it.

        Requests an allocation from current machine's resource manager and returns
        the job ID.

        :param nodes: sets the allocation size in terms of total nodes.
        :type nodes: int
        :param partition: the compute partition to use for the ensemble.
        :type partition: str
        :param bank: the bank to use for the ensemble, e.g. "wbronze". The default value
            of None allows the resource manager to choose the bank.
        :type bank: str
        :param name: the name to give the allocation.
        :type name: str
        :param timeout: the time limit to request for the allocation, in minutes.
        :type timeout: int
        :param repeats: the number of times to replicate the allocation if time expires
            but the ensemble is not yet complete.
        :param parallelism: if set to a positive integer N, reserve N cores for Themis
            itself, and run Themis in parallel on those cores. If you suspect Themis's
            performance is bottlenecking your ensemble, increase the parallelism.
        :type parallelism: int, optional
        :param early_stop: a positive integer indicating the number of minutes "early"
            that Themis should stop launching new runs. Once Themis has run for
            ``(timeout - early_stop)`` minutes, it will go to sleep until the
            allocation expires. While Themis is sleeping, no new runs will be launched,
            and updates to existing runs will be ignored. This option is generally
            only useful if an application actively checks the remaining time
            on an allocation.
        :type early_stop: int, optional
        :param allow_multiple: if ``True``, permit multiple concurrent executions
            (i.e. multiple simultaneous calls to ``execute_alloc`` or
            ``execute_local``). As this makes certain checks and behavior
            impossible, it is not the default.
        :type allow_multiple: bool, optional
        :param max_concurrency: the maximum number of concurrent runs Themis should
            allow at any given time. Too low a number may cause your ensemble to
            progress very slowly, much too large a number may eventually cause
            performance degradation on certain systems.
            If ``None``, pick a reasonable default value.
        :type max_concurrency: int, optional
        :returns: the job ID of the allocation
        """
        alloc = allocators.Allocation(
            nodes=nodes,
            partition=partition,
            bank=bank,
            name=name,
            timeout=timeout,
            repeats=repeats,
        )
        return self._execute(
            alloc,
            parallelism=parallelism,
            early_stop=early_stop,
            allow_multiple=allow_multiple,
            max_concurrency=max_concurrency,
        ).job_id

    def _execute(  # pylint: disable=too-many-arguments
        self,
        allocation,
        allocator=None,
        blocking=False,
        parallelism=None,
        early_stop=None,
        allow_multiple=False,
        max_concurrency=None,
        debug=False,
    ):
        if self.count_by_status(self.RUN_QUEUED) <= 0:
            raise RuntimeError("There are no runs to execute")
        resource_mgr = resource.identify_resource_manager(
            self._spec["resource_mgr"], path=self._spec["flux_path"]
        )
        if max_concurrency is None or max_concurrency < 0:
            max_concurrency = _DEFAULT_MAX_CONCURRENCY
        if allocator is None:
            allocator = resource_mgr.allocator()
        alloc_id = self._status_store.add_allocation(allocation)
        parallelism = parallelism if parallelism is not None else 0
        timeout = allocation.timeout if allocation.timeout is not None else 0
        early_stop = int(early_stop) if early_stop is not None else 0
        backend_commands = [
            resource_mgr.commands_to_launch_backend(
                timeout,
                parallelism,
                alloc_id,
                early_stop,
                allow_multiple,
                self._spec["setup_dir"],
                max_concurrency,
            )
        ]
        alloc_dir = resource.backend_dir(self._spec["setup_dir"], alloc_id)
        if not os.path.exists(alloc_dir):
            os.makedirs(alloc_dir)
        allocator.start(
            allocation, backend_commands, alloc_dir, alloc_dir,
        )
        if blocking and debug:
            allocator.wait(debug=debug)
        elif blocking:
            allocator.wait()
        return allocator

    def dry_run(self, *run_ids, **kwargs):
        """Perform one or more dry runs. If `run_ids` is empty, dry-run every run.

        Populate each run directory with
        the "run_symlink" and the "input deck" files, and call the
        application interface's ``prep_run`` function, if it exists.

        This is done in serial, so is safe to do on login nodes of HPC clusters.

        :param run_ids: the run IDs of the runs to dry-run.
        :param verbosity: If > 0, print messages about the progress of the dry runs.
        """
        verbosity = kwargs.get("verbosity", 0)
        if not run_ids:
            run_ids = self.filter_by_status(self.RUN_QUEUED)
        run_mapping = self._db.user_facing_run_info(run_ids)
        for run_id in run_ids:
            self._call_worker(run_id, run_mapping[run_id], "prep_run", verbosity)

    def _call_worker(self, run_id, run, app_interface_attr, verbosity):
        """Call a worker function directly, and return its return value."""
        backend.export_to_user_utils(self._spec, self._status_store, run_id, run)
        run_dir = utils.get_run_dir(run_id, self._spec["run_dir_names"], run.sample)
        if not os.path.exists(run_dir):
            os.makedirs(run_dir)
        app_interface_func = getattr(
            utils.import_app_interface(self._spec["app_interface"]),
            app_interface_attr,
            None,
        )
        if verbosity:
            message = "Executing {action} #{id} in {cwd!r}..."
            print(
                message.format(
                    action="dry run"
                    if app_interface_attr == "prep_run"
                    else app_interface_attr,
                    id=run_id,
                    cwd=os.path.relpath(run_dir),
                )
            )
        if app_interface_attr == "prep_run":
            return prepper.preparation(
                app_interface_func, run_dir, self._spec, run.sample, run.steps
            )
        if app_interface_attr == "post_run":
            return finisher.user_post_run(app_interface_func, run_dir)
        raise ValueError("Unrecognized interface attribute")

    def call_post_run(self, run_id):
        """Call the application interface's ``post_run`` function for a single run.

        :param run_id: the run ID of the run to execute ``post_run`` for.
        :type run_id: int
        :return: the return value of the ``post_run`` function. ``None`` if the
            function does not exist.
        """
        run_info = database.default_database(self._spec).get_run_info(run_id)
        if run_info is None:
            raise ValueError("No run has the id " + str(run_id))
        return self._call_worker(run_id, run_info.run, "post_run", 0,)

    def call_prep_ensemble(self):
        """Call the application interface's ``prep_ensemble`` function, if it exists."""
        self._call_ensemble_func("prep_ensemble")

    def call_post_ensemble(self):
        """Call the application interface's ``post_ensemble`` function, if it exists."""
        self._call_ensemble_func("post_ensemble")

    def _call_ensemble_func(self, user_func_name):
        user_func = getattr(
            utils.import_app_interface(self._spec["app_interface"]),
            user_func_name,
            None,
        )
        if not callable(user_func):
            return
        backend.export_to_user_utils(self._spec, self._status_store)
        if user_func_name == "prep_ensemble":
            backend.user_prep_ensemble(user_func, os.getcwd())
        elif user_func_name == "post_ensemble":
            backend.user_post_ensemble(user_func, os.getcwd())
        else:
            raise ValueError("Unrecognized interface attribute")

    def filter_by_status(self, *statuses):
        """Return an iterable of run IDs representing runs with a given status.

        The return value is not guaranteed to be perfectly accurate--there
        may be considerable delay in propagating a run's status change to
        the caller of this method.

        :param statuses: zero or more ``themis.Themis.RUN_*`` objects.
        :return: an iterable of integer run IDs, representing runs with
            one of the given statuses. If statuses is empty, return all run IDs.
        """
        return self._db.runs_with_completion_status(*statuses)

    def count_by_status(self, *statuses):
        """Return a count of runs with a given status.

        This method is equivalent to calling ``len()`` on the results
        of ``filter_by_status``; however, this method may be considerably
        faster for large ensembles.

        :param statuses: same meaning as in ``filter_by_status``.
        :return: the number of runs with one of the given statuses.
            If statuses is empty, return the total number of runs.
        """
        return self._db.count_runs_by_completion(*statuses)

    def progress(self):
        """Return a 2-tuple of ints representing the number of (completed, total) runs.

        Convenience wrapper around the ``count_by_status`` method.

        A completed run is one that has either succeeded, failed, or been killed.
        """
        completed = self.count_by_status(
            self.RUN_FAILURE, self.RUN_SUCCESS, self.RUN_KILLED
        )
        return (completed, self.count_by_status())

    def requeue_runs(self, run_ids, hard=False):
        """Requeue one or more completed runs.

        Restarted runs have their status reset to ``RUN_QUEUED``
        and will be executed at the earliest opportunity.
        If the run has not completed by the time this method is called,
        no action is taken.

        :param run_ids: an iterable of integer run IDs.
        :param hard: if ``True``, reset the given runs' progress back to step 0.
            If ``False``, maintain the runs' progress, but re-attempt execution.
        """
        self._db.mark_runs_to_restart(run_ids, hard)

    def restart_runs(self, *args, **kwargs):
        """Alias for ``requeue_runs``."""
        return self.requeue_runs(*args, **kwargs)

    def dequeue_runs(self, run_ids):
        """Dequeue one or more incomplete (i.e. queued) runs.

        Dequeued runs have their status set to ``RUN_KILLED`` after a delay.
        If the run has already finished when this method is called, no action is taken.
        If the run has not yet started, that run will never be started.

        :param run_ids: an iterable of integer run IDs.
        """
        self._db.mark_runs_to_kill(run_ids)

    def kill_runs(self, *args, **kwargs):
        """Alias for ``dequeue_runs``."""
        return self.dequeue_runs(*args, **kwargs)

    def add_runs(self, runs):
        """Add runs to the ensemble.

        The new runs will be executed at the earliest opportunity---either
        immediately or at the time of the next restart.

        :param runs: an iterable of ``Run`` objects.
        """
        runs = utils.validate_runs(
            runs,
            self._spec["run_dir_names"],
            application=self._spec["application"],
            batch_script=self._spec["app_is_batch_script"],
        )
        self._db.add_runs(runs)

    def runs(self, run_ids):
        """Get information about a set of runs.

        :param run_ids: an iterable of integer run IDs.
        :returns: a mapping from ``run_ids`` to ``Run`` objects. The objects will
            have additional ``result`` and ``status`` attributes.
        """
        return self._db.user_facing_run_info(run_ids)

    def as_dataframe(self, include_none=True):
        """Return a pandas DataFrame containing information about each run.

        The DataFrame will be indexed on run IDs, and have columns for
        each sample label.

        :param include_none: whether to include rows with no result
        :type include_none: bool, optional

        :raises ImportError: if ``pandas`` cannot be imported
        """
        from pandas import DataFrame

        mappings = self._db.as_mappings(include_none)
        for mapping in mappings:
            mapping.update(mapping["sample"])
            del mapping["sample"]
        dframe = DataFrame(mappings)
        dframe.set_index("run_id", inplace=True)
        dframe.drop(columns=["status"], inplace=True)
        return dframe

    def _info_to_write(self):
        mappings = self._db.as_mappings(True)
        for mapping in mappings:
            mapping["status"] = self.translate_enum(mapping["status"])
        return mappings

    def write_json(self, stream):
        """Write a JSON document containing a full set of data about the ensemble.

        :param stream: a file-like object that supports writing ``str``
        """
        runs = self._info_to_write()
        json.dump(runs, stream, indent=2, default=utils.Step.encode)

    def write_yaml(self, stream):
        """Write a YAML document containing a full set of data about the ensemble.

        :param stream: a file-like object that supports writing ``str``

        :raises ImportError: if ``yaml`` cannot be imported
        """
        import yaml

        runs = self._info_to_write()
        return yaml.dump(runs, stream)

    def write_csv(self, stream):
        """Write a CSV containing a full set of data about the ensemble.

        :param stream: a file-like object that supports writing ``str``

        :raises ValueError: if the results cannot be formatted into a csv
        """
        run_dicts = self._info_to_write()
        all_sample_labels = set()
        all_result_labels = set()
        for run_dict in run_dicts:
            all_sample_labels |= set(run_dict["sample"].keys())
            run_dict.update(run_dict["sample"])
            del run_dict["sample"]
            if isinstance(run_dict["result"], (numbers.Number, str)):
                run_dict["result"] = (run_dict["result"],)
            if isinstance(run_dict["result"], versions.Mapping):
                all_result_labels |= run_dict["result"].keys()
                run_dict.update(run_dict["result"])
                del run_dict["result"]
            elif isinstance(run_dict["result"], versions.Sequence):
                result_labels = [
                    "output_{}".format(i) for i in range(len(run_dict["result"]))
                ]
                for key, val in zip(result_labels, run_dict["result"]):
                    run_dict[key] = val
                del run_dict["result"]
                all_result_labels |= set(result_labels)
        fieldnames = ["run_id", "status", "tasks", "cores_per_task", "gpus_per_task"]
        fieldnames.extend(all_sample_labels)
        fieldnames.extend(sorted(all_result_labels))
        writer = csv.DictWriter(stream, fieldnames=fieldnames, extrasaction="ignore")
        writer.writeheader()
        for run_dict in run_dicts:
            writer.writerow(run_dict)

    def set_result(self, run_id, result):
        """Set the result for a run of the ensemble.

        :param run_id: the ID of the run
        :type run_id: int
        :param result: the result to add
        """
        self._db.add_result((run_id,), (result,))

    def run_dirs(self, nonexistent=False):
        """Return an iterable of (run_id, working_directory) pairs in arbitrary order.

        :param nonexistent: if ``True``, include directories that haven't yet been
            created.
        :type nonexistent: bool, optional
        """
        for mapping in self._db.as_mappings(True):
            run_dir = utils.get_run_dir(
                mapping["run_id"], self._spec["run_dir_names"], mapping["sample"]
            )
            if nonexistent or os.path.exists(run_dir):
                yield mapping["run_id"], run_dir

    def on_completion(self, args, cwd=os.curdir, stdout=None, stderr=None):
        """Provide an executable for Themis to launch once the ensemble has completed.

        The executable will be invoked during a call to ``Themis.execute_*`` once
        every run has finished.

        Only the most recent call to this function is remembered.

        :param args: an iterable of str defining the executable and its arguments,
            e.g. ``["/usr/bin/python", "-vvv", "myscript.py" "--foobar"]``.
        :param cwd: the working directory for the executable when it is launched
        :type cwd: str
        :param stdout: file to redirect stdout to;
            default redirects to a Themis log file
        :type stderr: str
        :param stderr: file to redirect stderr to;
            default redirects to a Themis log file
        :type stderr: str
        """
        if not all((isinstance(arg, str) for arg in args)):
            raise ValueError("arguments must be strings")
        if not os.path.isdir(cwd):
            raise ValueError("Not a directory: {!r}".format(cwd))
        executable = utils.which(args[0])
        if executable is None:
            raise ValueError("Cannot find executable {!r}".format(args[0]))
        args = list(args)  # make a shallow copy for modification
        args[0] = executable
        self._spec["on_completion"] = {
            "args": args,
            "cwd": os.path.abspath(cwd),
            "stdout": os.path.abspath(stdout) if stdout is not None else None,
            "stderr": os.path.abspath(stderr) if stderr is not None else None,
        }
        database.write_app_spec(self._spec, self._spec["setup_dir"])



"""
This subpackage contains part of Themis's backend.

The front end consists of the user-facing API and validation. The front end
launches the back end in an allocation, which does all of the work of actually
executing runs.

The backend consists of a number of worker scripts.

The worker scripts are designed to execute on compute nodes.
The prepper and finisher scripts are launched and managed by submitter scripts.
There can be multiple submitter scripts active in a single compute allocation
(i.e. each ensemble can have multiple submitters active at once),
and each of those submitters will have some number of preppers and finishers
active.
"""



"""Common functions and classes across the ensemble.backend package."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import logging
import logging.handlers
import os
import sys
import ssl
import cProfile
import pstats
import copy
import subprocess
import collections
import signal
import contextlib

from themis.utils import DirectoryManager
from themis import user_utils
from themis.versions import reprlib


PROFILE_ENV_VAR = "THEMIS_PROFILE"
LOGGER = logging.getLogger(__name__)


class SignalCaught(Exception):
    """Exception raised on catching a signal"""

    def __init__(self, message, signum):
        super(SignalCaught, self).__init__(message)
        self.signum = signum


def signal_handler_exit(signum, _):
    """Signal handler for `signal` module."""
    LOGGER.info("Signal handler called with signal %s", signum)
    raise SignalCaught("Signal handler called with signal {}".format(signum), signum)


def handle_signals():
    """Attach a signal handler that raises an exception on SIGALRM and SIGTERM."""
    for signum in (signal.SIGALRM, signal.SIGTERM):
        signal.signal(signum, signal_handler_exit)


@contextlib.contextmanager
def managed_shutdown(thing, *args, **kwargs):
    """Invoke `shutdown` on `thing` upon completion of a block of code."""
    try:
        yield thing
    finally:
        thing.shutdown(*args, **kwargs)


def user_prep_ensemble(prep_ensemble, root_dir):
    """Call user's prep_ensemble function and return None.

    Any return value from the prep_ensemble function is ignored.

    :param prep_ensemble: the user's prep_ensemble function, a callable
    """
    with DirectoryManager(root_dir):
        try:
            prep_ensemble()
        except:  # pylint: disable=bare-except
            LOGGER.exception(
                "The following exception was generated "
                "by the prep_ensemble function:"
            )


def user_post_ensemble(post_ensemble, root_dir):
    """Call the user's post_ensemble. Return True if new points were added.

    :param post_ensemble: the user's post_ensemble function, a callable
    """
    # make sure we stay in the right directory
    LOGGER.info("Calling user's post_ensemble")
    with DirectoryManager(root_dir):
        try:
            post_ensemble()
        except:  # pylint: disable=bare-except
            LOGGER.exception(
                "The following exception was generated "
                "by the post_ensemble function:"
            )


def invoke_completion_script(on_completion_dict):
    """Invoke a user's completion script."""
    stdout = None
    stderr = None
    LOGGER.info("Launching completion script")
    try:
        if on_completion_dict["stdout"] is not None:
            stdout = open(on_completion_dict["stdout"], "a")
        if on_completion_dict["stderr"] is not None:
            stderr = open(on_completion_dict["stderr"], "a")
        proc = subprocess.Popen(
            on_completion_dict["args"],
            stdout=stdout,
            stderr=stderr,
            cwd=on_completion_dict["cwd"],
            universal_newlines=True,
        )
    except (OSError, FileNotFoundError):
        LOGGER.exception("Couldn't launch on_completion application:")
    else:
        # check to ensure process exited cleanly
        if proc.wait() != 0:
            LOGGER.error(
                "on_completion script exited with returncode %i", proc.returncode
            )
    finally:
        # close the IO streams
        if stdout is not None:
            stdout.close()
        if stderr is not None:
            stderr.close()


def profile(profiler, func, *args, **kwargs):
    """Profile a given function if PROFILE_ENV_VAR is set."""
    if os.getenv(PROFILE_ENV_VAR, ""):
        profiler.enable()
        return_val = func(*args, **kwargs)
        profiler.disable()
        return return_val
    return func(*args, **kwargs)


def get_profiler():
    """Return a Profile instance or None."""
    if os.getenv(PROFILE_ENV_VAR, ""):
        return cProfile.Profile()
    return None


def profile_from_new(func, *args, **kwargs):
    """Profile a function with a new profile """
    profiler = None
    if os.getenv(PROFILE_ENV_VAR, ""):
        profiler = cProfile.Profile()
    return (profiler, profile(profiler, func, *args, **kwargs))


def dump_profile_data(path, *profilers):
    """Dump profile data accumulated by `profilers` to a file at `path`."""
    if os.getenv(PROFILE_ENV_VAR, "") and profilers:
        profile_stats = pstats.Stats(profilers[0])
        for additional_profile in profilers[1:]:
            if additional_profile is not None:
                profile_stats.add(additional_profile)
            profile_stats.dump_stats(path)


def clear_user_utils():
    """Clear all information exported to the user_utils module."""
    # pylint: disable=protected-access
    user_utils._RUN_ID = None
    user_utils._STATUS_DB = None
    user_utils._SETUP_DIR = None
    user_utils._RESOURCES = None


def export_to_user_utils(app_spec, status_db, run_id=None, run=None):
    """Export the given information about this run to the user_utils module."""
    # pylint: disable=protected-access
    user_utils._RUN_ID = run_id
    user_utils._STATUS_DB = status_db
    user_utils._SETUP_DIR = app_spec["setup_dir"]
    user_utils._RUN = copy.deepcopy(run)


def _make_stream_handlers(streams):
    """Make logging handlers for writing to a stream."""
    formatter = logging.Formatter("THEMIS - %(levelname)s - %(asctime)s - %(message)s")
    handlers = []
    for stream in streams:
        stream_handler = logging.StreamHandler(stream)
        stream_handler.setFormatter(formatter)
        handlers.append(stream_handler)
    return handlers


def _make_queue_handlers(streams):
    """Make a QueueHandler for writing log messages asynchronously.

    Only available in Python 3.
    """
    import queue

    que = queue.Queue()
    queue_handler = logging.handlers.QueueHandler(que)
    stream_handlers = _make_stream_handlers(streams)
    listener = logging.handlers.QueueListener(
        que, *stream_handlers, respect_handler_level=True
    )
    listener.start()
    return (listener, [queue_handler])


def _make_handlers(streams):
    """Return a log handler."""
    if sys.version_info.major >= 3:
        return _make_queue_handlers(streams)
    return (None, _make_stream_handlers(streams))


def _setup_logger(streams):
    """Redirect stderr, stdout, and the root logger to log_file."""
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    listener, handlers = _make_handlers(streams)
    for handler in handlers:
        root_logger.addHandler(handler)
    return listener


@contextlib.contextmanager
def managed_logging(*streams):
    """Manage logging and shut down on exit."""
    listener = _setup_logger(streams)
    try:
        yield
    finally:
        if listener is not None:
            listener.stop()


class RunFeeder(object):
    """Instances of this class produce RunInfo objects representing new runs.

    :param ensemble_db: the EnsembleDatabase object
        to use for fetching new runs.
    :param cache_size: the size of the cache for RunInfo objects.
        A larger cache means less-frequent but larger trips to
        the database, but a smaller cache makes for better
        load-balancing among multiple RunFeeder instances.
    """

    def __init__(self, ensemble_db, cache_size, enable_runs=False):
        """Initialize a new instance."""
        self._ensemble_db = ensemble_db
        self._run_cache = collections.deque()
        self._disabled_run_ids = set()
        self._enable_runs = enable_runs
        self._cache_max_size = cache_size
        self._last_fill_attempt_succeeded = True
        self.fill_cache()

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}({}, {})".format(
            type(self).__name__, self._ensemble_db, self._cache_max_size
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()
        return False

    def new_runs(self, limit):
        """Return new runs and their statuses.

        Returns a two-tuple; the first entry (index 0) contains a sequence of
        run IDs identifying the new runs, the second contains the statuses
        corresponding to those IDs.

        :param limit: the max number of runs to return.
        """
        if limit <= 0:
            # return an empty tuple
            return ()
        return [
            self._run_cache.popleft() for _ in range(min(limit, len(self._run_cache)))
        ]

    def poll_cache(self, fill_factor=0.25):
        """Fill the cache if it is less than `fill_factor` full.

        If the last attempt at fetching runs from the database failed,
        don't fetch any runs.
        """
        if self._last_fill_attempt_succeeded and len(self._run_cache) < (
            fill_factor * self._cache_max_size
        ):
            self.fill_cache()

    def fill_cache(self):
        """Fill up the cache of RunInfo objects.

        Return True if the fill attempt succeeds, False other if it fails
        (because there are no eligible runs in the database).
        """
        runs_to_add = self._cache_max_size - len(self._run_cache)
        if runs_to_add < 1:
            return True
        new_runs = self._ensemble_db.new_runs(limit=runs_to_add)
        self._run_cache.extend(new_runs)
        if self._enable_runs:
            for run in new_runs:
                self._disabled_run_ids.add(run.run_id)
        if not new_runs:
            self._last_fill_attempt_succeeded = False
        else:
            self._last_fill_attempt_succeeded = True
        return self._last_fill_attempt_succeeded

    def is_empty(self):
        """Return True if all new runs have been exhausted, False otherwise."""
        return len(self._run_cache) <= 0

    def shutdown(self):
        """Enable all runs still held in the cache."""
        if self._disabled_run_ids:
            LOGGER.debug("Enabling runs %s", reprlib.repr(self._disabled_run_ids))
            self._ensemble_db.enable_runs(self._disabled_run_ids)



"""This module provides a python API and command-line
interface to submit a batch script repeatedly.

The API consists of the ``BatchSubmitter`` class.

Run the module as a script to produce the command-line interface.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import argparse

import themis
from themis import utils
from themis import resource
from themis.backend.worker.prepper import populate_run_dir

class BatchSubmitter(object):  # pylint: disable=too-many-instance-attributes
    """Instances of this class are used to create and launch ensembles.

    The ensemble proceeds by submitting a series of batch scripts to a
    resource manager. Each run of the ensemble is a new allocation (a new "batch job")
    for the resource manager, so each run will have to wait in the batch queue.

    Before the batch scripts are launched, parse them for variables declared with
    "%%".

    :param batch_script: the path to the batch script to execute. The
        batch script will be hard-copied into each run directory, parsed,
        and then executed.
    :type batch_script: str
    :param resource_mgr: the resource manager to submit the script to.
        Accepted values are ``"slurm"``, ``"moab"``, ``"lsf"``, and ``"flux"``.
    :type resource_mgr: str
    :param samples: an iterable of mappings defining the sample for each run.
        The number of samples is equal to the number of runs; furthermore, each
        sample is assigned a unique "run ID" corresponding to its position
        in the ``samples`` iterable.
    :param run_parse: A file path or iterable of file paths. The files specified
        will be hard-copied into the run directories and parsed;
        see :ref:`here <text_replacement>` for more.
    :type run_parse: str or iterable of str, optional
    :param run_copy: A file path or iterable of file paths. Unix-style path
        patterns are supported as well. The files/directories
        specified will be hard-copied into the run directories. ``None``, the
        default, specifies that no files will be copied.
    :type run_copy: str or iterable of str, optional
    :param run_symlink: A file path or iterable of file paths. Unix-style path
        patterns are supported as well. The files/directories
        specified will be symlinked into the run directories. ``None``, the
        default, specifies that no files will be symlinked.
    :type run_symlink: str or iterable of str, optional
    :param run_dir_names: The file system paths of the run directories.
        This argument should be a `python format string
        <https://docs.python.org/3/library/string.html#formatstrings>`_, where
        the field names correspond to the names of the variables in the samples
        argument. For instance, if the variables in the samples argument are
        "hydrostatics" and "viscocity", you might pass in the string
        ``"hydro={hydrostatics}/visc={viscocity}"``. This string will be
        formatted each run to yield the run directory; so one directory
        might be ``hydro=17.6/visc=35``. Note that
        the posix directory separator character "/"
        in the example string means that the resulting run directory
        will be in fact a sequence of two directories.
        The default value of ``None`` lets the naming scheme be determined internally.
    :type run_dir_names: str, optional
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        batch_script,
        samples,
        resource_mgr,
        run_parse=None,
        run_copy=None,
        run_symlink=None,
        run_dir_names=None,
    ):
        """Initialize a new instance."""
        self._batch_script = utils.validate_application(batch_script)
        self._resource_mgr = resource.identify_resource_manager(resource_mgr)
        self._allocator = self._resource_mgr.allocator()
        run_parse, run_copy, run_symlink = utils.validate_run_files(
            run_parse, run_copy, run_symlink
        )
        run_parse.append(batch_script)
        self._run_parse = run_parse
        self._run_copy = run_copy
        self._run_symlink = run_symlink
        if run_dir_names is None:
            self._run_dir_names = os.path.abspath(utils.DEFAULT_RUN_DIR_NAMES)
        else:
            self._run_dir_names = os.path.abspath(run_dir_names)
        self._samples = utils.validate_samples(
            samples, run_dir_names, check_nonempty=True
        )

    def __repr__(self):
        return "{}({!r}, {}, {})".format(
            type(self).__name__, self._batch_script, self._resource_mgr, self._samples
        )

    def execute(self, verbosity=0):
        """Submit the batch scripts to the resource manager.

        :returns: the job IDs of the submitted batch scripts.
        :param verbosity: If > 0, print messages about the progress of the submissions.
        """
        job_ids = []
        for i in range(len(self._samples)):
            run_dir = self._dry_run_one(i, verbosity)
            parsed_batch_script = os.path.join(
                run_dir, os.path.basename(self._batch_script)
            )
            jobid = self._allocator.launch_batch_script(parsed_batch_script, run_dir)
            if verbosity > 0:
                print(
                    "Submitted batch script, job ID is {}".format(
                        self._allocator.job_id
                    )
                )
            job_ids.append(jobid)
        return job_ids

    def dry_run(self, *run_ids, **kwargs):
        """Populate run directories with the `run_*` files.

        That is, symlink the `run_symlink` files, and so on.

        :param run_ids: the run IDs of the runs to dry-run.
        :param verbosity: If > 0, print messages about the progress of the dry runs.
        """
        verbosity = kwargs.get("verbosity", 0)
        if not run_ids:
            run_ids = range(len(self._samples))
        for run_id in run_ids:
            self._dry_run_one(run_id, verbosity)

    def _dry_run_one(self, run_id, verbosity):
        """Dry run a single run.

        :returns: the path to the run directory.
        """
        sample = self._samples[run_id]
        run_dir = utils.get_run_dir(run_id, self._run_dir_names, sample)
        if not os.path.exists(run_dir):
            os.makedirs(run_dir)
        if verbosity > 0:
            print(
                "Populating {!r} for run #{}...".format(
                    os.path.relpath(run_dir), run_id
                )
            )
        populate_run_dir(
            self._run_symlink, self._run_copy, self._run_parse, sample, run_dir
        )
        return run_dir


def setup_parser():
    """Setup the command-line parser."""
    parser = argparse.ArgumentParser(
        description=(
            "Submit a batch script to a resource manager repeatedly, modifying it "
            "each time. This tool is a simpler offshoot of Themis with very different "
            "behavior. Themis runs inside of a single allocation and manages each run "
            "individually. This tool, `themis-laf`, just submits a collection of "
            "batch scripts, and then quits---nothing more."
        )
    )
    parser.add_argument(
        "batchscript", type=str, help="Path to the batch script to modify and submit"
    )
    utils.add_parameterfile_arg(parser)
    parser.add_argument(
        "resourcemgr",
        type=lambda x: str(x).lower(),
        choices=["slurm", "moab", "lsf", "sbatch", "none"],
        help="Resource manager to submit the script to",
    )
    utils.add_common_creation_args(parser)
    parser.add_argument(
        "-d",
        "--dry-run",
        type=int,
        nargs="*",
        help=(
            "Perform dry runs: don't submit the batch scripts, but set everything "
            "up for submission. Pass one or more integers to perform specific dry "
            "runs. (Each run is identified by its 0-based row in the CSV.) "
            "Default behavior is to dry-run every run."
        ),
        default=None,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity of output",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + str(themis.__version__)
    )
    return parser.parse_args()


def main():
    """Run the command-line interface to the ``BatchSubmitter``."""
    args = setup_parser()

    print(args)
    samples = utils.read_csv(args.parameterfile)
    mgr = BatchSubmitter(
        args.batchscript,
        samples,
        args.resourcemgr,
        args.run_parse,
        args.run_copy,
        args.run_symlink,
        args.run_dir_names,
    )
    if args.dry_run is not None:
        mgr.dry_run(*args.dry_run)
    else:
        job_id_strs = [str(job_id) for job_id in mgr.execute()]
        print("Batch job IDs are: " + ", ".join(job_id_strs))


if __name__ == "__main__":
    main()



"""
This module defines classes and functions related to creating HPC allocations.

Instances of the Allocator class interface with resource managers
(e.g. LSF on LC's CORAL systems) to create allocations.

Allocations are sometimes also known as "jobs"; the applications launched inside
them (via, e.g., `srun`) are called "job steps", although the terminology is used
inconsistently.

Instances of the Allocation class are used to describe an allocation to
an Allocator (which then creates that allocation).

The ShellScript class is used to write batch scripts. All the supported
resource managers only accept jobs (allocations) in the form of scripts.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import subprocess as sp
import time
import os
import stat

from themis import utils
from themis.versions import basestring


class ShellScript(object):
    """Create an object representation of a shell script."""

    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.path, self.shell)

    def __init__(self, path, shell="bash"):
        """Create an object representing a new shell script.

        :param path: the path to the new shell script.
        :param shell: the shell to use. Only affects the shebang.
        """
        self.path = path
        self.shell = shell
        self.headers = []
        self.commands = []
        self.working_directory = None

    def write(self):
        """Write the shell script to the file system.

        Ensures the file has execute permissions.
        Returns the path of the written file.
        """
        with open(self.path, "w") as file_handle:
            file_handle.write("#!" + utils.which(self.shell) + "\n")
            for header in self.headers:
                file_handle.write(header + "\n")
            file_handle.write("\n")
            if self.working_directory is not None:
                file_handle.write("cd " + self.working_directory + "\n")
            for command in self.commands:
                file_handle.write(command + "\n")
        # ensure usr has r/w/x permisions
        os.chmod(self.path, stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR)
        return os.path.abspath(self.path)


def _get_launch_script_name(script_dir):
    """Return a path to the ensemble's allocation script."""
    if script_dir is None:
        script_dir = os.getcwd()
    return os.path.join(script_dir, "ensemble_launch_script.sh")


class Allocation(object):  # pylint: disable=too-few-public-methods
    """Represents an allocation on a HPC system.

    :param nodes: sets the allocation size for the ensemble in terms of total nodes.
    :type nodes: int
    :param partition: the compute partition to use for the ensemble.
    :type partition: str
    :param bank: the bank to use for the ensemble, e.g. "wbronze". The default value
        of None allows the resource manager to choose the bank.
    :type bank: str
    :param name: the name to give the allocation.
    :type name: str
    :param timeout: the time limit to request for the allocation, in minutes.
    :type timeout: int
    :param repeats: the number of times to replicate the allocation if time expires
        but the ensemble is not yet complete
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        nodes=1,
        partition=None,
        bank=None,
        name="themis",
        timeout=None,
        repeats=0,
    ):
        self.partition = utils.type_check(partition, basestring, type(None))
        self.bank = utils.type_check(bank, basestring, type(None))
        self.name = utils.type_check(name, basestring)
        self.repeats = repeats
        if timeout is not None:
            utils.range_check(timeout, min_val=0, name="timeout")
        self.timeout = timeout
        self.nodes = utils.range_check(nodes, min_val=1, name="nodes")

    def __repr__(self):
        """Return str representation of constructor call."""
        return "{}(nodes={}, partition={!r}, bank={!r}, name={!r}, timeout={})".format(
            type(self).__name__,
            self.nodes,
            self.partition,
            self.bank,
            self.name,
            self.timeout,
        )


class Allocator(object):
    """Abstract base class; defines interface for interactions with resource managers.

    Allocator classes are designed to request an allocation of resources from the
    system, and then run a designated script within that allocation.

    Allocators for HPC resources generally build a batch submission script, then pass it
    to the resource manager for that system. Sometimes they may launch a command
    that requests an interactive allocation.

    Allocators for generic computers are usually trivial, because in that case the way
    to request an 'allocation of resources' for a set of programs is just to launch
    those programs.
    """

    def __repr__(self):
        """Return string representation of constructor call"""
        return "{}()".format(type(self).__name__)

    def start(self, allocation, applications, script_dir=None, working_dir=None):
        """Request an allocation of resources and launch an application within it.

        :param allocation: an Allocation instance representing the
            allocation to request.
        :param applications: the applications to launch within the allocation. This
            argument should be an iterable of str,
            e.g. ["cd /my/dir", "srun -n16 my/app", "cat app_results | grep hydro"]
        :param script_dir: the directory in which to write the allocation script
            to write (if any). Should be ignored if not applicable
        """

    def wait(self, sleep_interval=10):
        """Wait for the ensemble to complete. For use by testing infrastructure."""
        raise NotImplementedError()


class BatchScriptAllocator(Allocator):
    """Abstract base class for all Allocators that write batch scripts."""

    def __init__(self, job_id=None):
        """Constructor"""
        self.job_id = job_id

    def start(self, allocation, applications, script_dir=None, working_dir=None):
        if working_dir is None:
            working_dir = os.getcwd()
        script_path = self._write_batch_script(
            allocation, applications, script_dir, working_dir
        )
        self.job_id = self.launch_batch_script(script_path, working_dir)
        return self.job_id

    def _write_batch_script(self, allocation, applications, script_dir, working_dir):
        """Write a batch script."""
        raise NotImplementedError()

    def launch_batch_script(self, script_path, working_dir):
        """Submit a batch script and return the job ID."""
        raise NotImplementedError()


class MoabAllocator(Allocator):
    """A partially-implemented allocator.

    Moab is not a supported resource manager. However,
    the launch_batch_script method is needed.
    """

    @staticmethod
    def launch_batch_script(script_path, working_dir):
        """Submit a batch script to Moab; return the job ID."""
        subproc = sp.Popen(
            ["msub", script_path],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            universal_newlines=True,
            cwd=working_dir,
        )
        stdout, stderr = subproc.communicate()
        if subproc.returncode != 0:
            raise Exception(
                "Couldn't launch Moab batch script, "
                "msub exited with returncode {}:\n{}".format(subproc.returncode, stderr)
            )
        return int(stdout)

    def start(self, allocation, applications, script_dir=None, working_dir=None):
        raise NotImplementedError()


class SbatchAllocator(BatchScriptAllocator):
    """A Allocator for the Slurm resource manager.

    Launches by creating a sbatch script, the submitting it to Slurm by running
    sbatch as a subprocess. Sbatch adds the job to the queue, returns a job ID,
    and exits immediately.
    """

    @classmethod
    def _write_batch_script(cls, allocation, applications, script_dir, working_dir):
        """Build a Slurm batch script."""
        launch_script = ShellScript(_get_launch_script_name(script_dir))
        launch_script.working_directory = working_dir
        declaration = "#SBATCH"
        standard_slurm_headers = [
            "{} -J {}".format(declaration, allocation.name),
            "{} -N {}".format(declaration, allocation.nodes),
            "{} --exclusive".format(declaration),
        ]
        launch_script.headers.extend(standard_slurm_headers)
        if allocation.bank is not None:
            launch_script.headers.append(
                "{} -A {}".format(declaration, allocation.bank)
            )
        if allocation.partition is not None:
            launch_script.headers.append(
                "{} -p {}".format(declaration, allocation.partition)
            )
        if allocation.timeout is not None:
            launch_script.headers.append(
                "{} -t {}".format(declaration, int(allocation.timeout))
            )
        launch_script.commands.extend(applications)
        launch_script.commands.append("wait")
        return launch_script.write()

    @staticmethod
    def launch_batch_script(script_path, working_dir):
        """Submit a Sbatch batch script; return the jobid."""
        subproc = sp.Popen(
            ["sbatch", script_path],
            stdout=sp.PIPE,
            stderr=sp.STDOUT,
            universal_newlines=True,
            cwd=working_dir,
        )
        stdout, _ = subproc.communicate()
        if subproc.returncode != 0:
            raise Exception(
                "Couldn't launch Slurm batch script, Sbatch exited with "
                "returncode {}:\n{}".format(subproc.returncode, stdout)
            )
        return int(stdout.split()[-1])

    def wait(self, sleep_interval=5):
        while not self._done():
            time.sleep(sleep_interval)

    def _done(self):
        """Return True if ensemble job is done, False otherwise.

        Runs the sacct command to see the if the ensemble is done.
        """
        subproc = sp.Popen(
            ["sacct", "-X", "-j", str(self.job_id)],
            stdout=sp.PIPE,
            universal_newlines=True,
        )
        stdout, _ = subproc.communicate()
        job_completed = ("COMPLETED", "CANCELLED", "TIMEOUT", "FAILED")
        for status in job_completed:
            if status in stdout:
                return True
        return False


class BsubAllocator(BatchScriptAllocator):
    """Allocator for the LSF resource manager.

    Launches by creating a bsub script, the submitting it to LSF by running
    bsub as a subprocess. Bsub adds the job to the queue, returns a job ID,
    and exits immediately.
    """

    @classmethod
    def _write_batch_script(cls, allocation, applications, script_dir, working_dir):
        """Write a Bsub batch script."""
        launch_script = ShellScript(_get_launch_script_name(script_dir))
        launch_script.working_directory = working_dir
        declaration = "#BSUB"
        standard_lsf_headers = [
            "{} -J {}".format(declaration, allocation.name),
            "{} -nnodes {}".format(declaration, allocation.nodes),
        ]
        launch_script.headers.extend(standard_lsf_headers)
        if allocation.bank is not None:
            launch_script.headers.append(
                "{} -G {}".format(declaration, allocation.bank)
            )
        if allocation.partition is not None:
            launch_script.headers.append(
                "{} -q {}".format(declaration, allocation.partition)
            )
        if allocation.timeout is not None:
            launch_script.headers.append(
                "{} -W {}".format(declaration, int(allocation.timeout))
            )
        launch_script.commands.extend(applications)
        launch_script.commands.append("wait")
        return launch_script.write()

    @staticmethod
    def launch_batch_script(script_path, working_dir):
        """Submit a Bsub batch script to LSF; return the jobid."""
        with open(script_path, "r") as file_handle:
            subproc = sp.Popen(
                ["bsub"],
                stdin=file_handle,
                stdout=sp.PIPE,
                stderr=sp.PIPE,
                universal_newlines=True,
                cwd=working_dir,
            )
        stdout, stderr = subproc.communicate()
        if subproc.returncode != 0:
            raise Exception(
                "Couldn't launch LSF batch script; "
                "Bsub exited with returncode {}:\n{}".format(subproc.returncode, stderr)
            )
        return int(stdout.split()[1].strip("><"))

    def wait(self, sleep_interval=5):
        while not self._done():
            time.sleep(sleep_interval)

    def _done(self):
        """Return True if the ensemble is done, False otherwise.

        Runs bjobs as a subprocess and checks if the job ID is in the 'done'
        category of information printed to stdout.
        """
        subproc = sp.Popen(
            ["bjobs", "-d"], stdout=sp.PIPE, stderr=sp.PIPE, universal_newlines=True
        )
        stdout, _ = subproc.communicate()
        return str(self.job_id) in stdout


class InteractiveAllocator(BatchScriptAllocator):
    """An allocator for a generic computer with no special allocation system.

    This allocator works on any system, including laptops, but
    it is a bad choice for production runs on HPC systems.
    No allocation is requested--the scheduler is simply submitted as a subprocess.
    However, for running interactively on HPC systems (either within an allocation
    or on a login node), this Allocator should work well.
    """

    def __init__(self, job_id=None):
        """Constructor"""
        super(InteractiveAllocator, self).__init__(job_id)
        self._proc_handle = None

    @classmethod
    def _write_batch_script(cls, _, applications, script_dir, working_dir):
        """Write a script to execute applications. The allocation is ignored."""
        script = ShellScript(_get_launch_script_name(script_dir))
        script.commands.extend(applications)
        script.commands.append("wait")
        working_dir = working_dir if working_dir is not None else os.getcwd()
        script.working_directory = working_dir
        return script.write()

    def launch_batch_script(self, script_path, working_dir):
        """Launch the script as a subprocess; return the PID."""
        stdio_path = os.path.join(working_dir, "themis_interactive_script.log")
        with open(stdio_path, "a") as stdio:
            self._proc_handle = sp.Popen(
                [script_path], cwd=working_dir, stdout=stdio, stderr=sp.STDOUT
            )
        return self._proc_handle.pid

    def wait(self, sleep_interval=10, debug=False):  # pylint: disable=arguments-differ
        """Wait for the subprocess to complete, then return"""
        if self._proc_handle is not None:
            if self._proc_handle.wait() != 0 and debug:
                raise RuntimeError("Wait returned nonzero")
        else:
            os.waitpid(self.job_id, os.WNOHANG)

    def done(self):
        """Return True if the subprocess is done, False otherwise"""
        if self._proc_handle is not None:
            return self._proc_handle.poll() is not None
        return not self._check_pid(self.job_id)

    @classmethod
    def _check_pid(cls, pid):
        """Return true if a process given by pid has completed."""
        try:
            os.kill(pid, 0)
        except OSError:
            return True
        else:
            return False


class FluxAllocator(Allocator):
    def __init__(self, job_id=None):
        self.job_id = None

    def start(self, allocation, applications, script_dir=None, working_dir=None):
        args = [
            "flux",
            "mini",
            "batch",
            "--wrap",
            "--exclusive",
            "-N",
            str(allocation.nodes),
        ]
        if allocation.name is not None:
            args.append("--job-name=" + str(allocation.name))
        if allocation.timeout is not None:
            args += ["-t", str(allocation.timeout) + "m"]
        separated_applications
        for app in applications:
            separated_applications.append(app + ";")
        proc = sp.Popen(
            args + applications + ["wait"],
            stdout=sp.PIPE,
            stderr=sp.STDOUT,
            universal_newlines=True,
            cwd=working_dir,
        )
        stdout, _ = proc.communicate()
        if proc.returncode != 0:
            raise Exception(
                "Couldn't launch Flux job, flux mini batch exited with "
                "returncode {}:\n{}".format(subproc.returncode, stdout)
            )
        return stdout

    def wait(self, sleep_interval=10):
        """Wait for the ensemble to complete. For use by testing infrastructure."""
        proc = sp.Popen(
            ["flux", "job", "attach", str(self.job_id)],
            stdout=sp.DEVNULL,
            stderr=sp.PIPE,
            universal_newlines=True,
        )
        _, stderr = proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError(stderr)



"""Utility functions for the backend of themis"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import argparse
import logging
import os

from themis import database
from themis import backend
from themis.backend.network import CERT_NAME
from themis.backend.worker import network
import themis.utils


def profile_worker(main_func, worker):
    """This function provides a layer on top of main(), to allow for profiling main().

    In order to properly direct the profile output, need access to the run ID.
    """
    with backend.managed_logging(sys.stderr):
        run_id, server_args, setup_dir = read_command_line("prep an application run")
        profiler, _ = backend.profile_from_new(
            main_func, run_id, server_args, setup_dir
        )
        if run_id < 1000:
            backend.dump_profile_data(
                "submitter_profile_stats_{}_{}.pstat".format(worker, run_id), profiler
            )


def read_command_line(script_description):
    """Parse worker command line arguments and return them.

    :param script_description: the description to display for --help messages
    """
    parser = argparse.ArgumentParser(description=script_description)
    parser.add_argument(
        "run_id", metavar="N", type=int, help="the (unique) id of this application run"
    )
    parser.add_argument("setup_dir", help="the setup dir for the ensemble")
    parser.add_argument(
        "--server-args",
        metavar="ARG",
        type=str,
        nargs="*",
        help="arguments to connect to the server (if it exists)",
    )
    args = parser.parse_args()
    return (args.run_id, args.server_args, args.setup_dir)


def setup_script(run_id, server_args, setup_dir):
    """Set up a worker script.

    Connects to the ensemble database, collects the sample and app spec, imports the
    application interface, and creates the run directory. Returns all the
    collected information.

    :param run_id: ID identifying the worker
    :param server_args: a sequence of arguments used to connect to the ensemble server
    """
    # setup the logger to log into the new directory, and direct output to the log file
    logger = logging.getLogger(__name__)
    logger.info("Server connection information: %s", server_args)
    if not server_args:
        app_spec = database.get_app_spec(setup_dir)
        ensemble_db = database.default_database(app_spec)
        run_info = ensemble_db.get_run_info(run_id)
    else:
        ensemble_db = network.WorkerClient(
            os.path.join(setup_dir, CERT_NAME), *server_args
        )
        app_spec, run_info = ensemble_db.get_worker_setup(run_id)
    sample = run_info.sample
    # import the application interface
    app_interface = themis.utils.import_app_interface(app_spec["app_interface"])
    run_dir = themis.utils.get_run_dir(run_id, app_spec["run_dir_names"], sample)
    logger.info("Using database: %s", ensemble_db)
    status_db = database.get_status_store()
    backend.export_to_user_utils(app_spec, status_db, run_id, run_info.run)
    return (ensemble_db, app_interface, run_dir, app_spec, run_info)



"""This module provides the primary interface to resource-manager-specific code.

Aside from the ResourceManager class itself, which defines the primary interface,
there are also a handful of utility functions in this module.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import os
import math
import time
import socket
import subprocess
import platform

from themis.resource import executors
from themis.resource import allocators
from themis import utils


_UNIDENTIFIED_RMGR = None

_RESOURCE_MANAGERS = {}

_BACKEND_PATH = os.path.join(os.path.abspath(__file__).rsplit(os.sep, 2)[0], "backend")

_WORKER_PATH = os.path.join(
    os.path.abspath(__file__).rsplit(os.sep, 2)[0], "backend", "worker"
)


def list_resource_mgr_identifiers():
    """List the identifiers of supported resource managers."""
    return list(key for key in _RESOURCE_MANAGERS if key != Moab.identifier)


def valid_resource_mgr(identifier):
    """Return True if `identifier` identifies a supported resource manager."""
    if isinstance(identifier, str):
        identifier = identifier.lower()
    return identifier in _RESOURCE_MANAGERS


def _resource_mgr_from_env():
    """Attempt to identify the resource manager from the environment.

    Return the empty string if no match is found
    """
    sys_type = os.getenv("SYS_TYPE", "")
    if sys_type == "" and platform.system().lower() in ["darwin", "windows"]:
        return NoResourceManager.identifier
    if sys_type.startswith("toss") or (utils.which("srun") and utils.which("sbatch")):
        if os.path.isdir("/opt/cray"):
            return LANLSlurm.identifier
        return Slurm.identifier
    if sys_type.startswith("blueos") or (utils.which("bsub") and utils.which("jsrun")):
        return Lsf.identifier
    # check Flux last because it can run inside the others
    if (
        os.getenv("FLUX_URI") is not None
        or subprocess.Popen(["flux", "resource", "list"]).wait() == 0
    ):
        return Flux.identifier
    return _UNIDENTIFIED_RMGR


def default_resource_manager_id():
    """Return the current type of system. Return values are one of None,
    'toss', 'blueos', 'flux' and 'suse-linux'.

    Note that None is actually an identifier for a resource manager,
    the NoResourceManager.

    Raises NotImplementedError if system is not recognized.
    """

    rsrc_mgr = _resource_mgr_from_env()
    if rsrc_mgr == _UNIDENTIFIED_RMGR:
        not_supported_msg = (
            "Your machine {!r} is not recognized, and may not be supported."
        )
        raise NotImplementedError(not_supported_msg.format(socket.gethostname()))
    return rsrc_mgr


def identify_resource_manager(identifier=_UNIDENTIFIED_RMGR, path=None):
    """Return the resource manager denoted by the given identifier.

    If identifier is None, get the default resource manager for this machine.
    """
    if identifier == _UNIDENTIFIED_RMGR:
        identifier = default_resource_manager_id()
    elif isinstance(identifier, str):
        identifier = identifier.lower()
    try:
        return _RESOURCE_MANAGERS[identifier](path)
    except KeyError:
        raise ValueError("No resource manager recognized by {!r}".format(identifier))


def _register_rmgr(cls):
    """Register a ResourceManager subclass with the identify_resource_manager func."""
    _RESOURCE_MANAGERS[cls.identifier] = cls
    return cls


def validate_flux_path(use_flux):
    """Search a couple of predefined paths for a Flux installation."""
    if isinstance(use_flux, str):
        path = utils.which(use_flux)
        if path is None:
            raise ValueError(
                "Can't find Flux installation at user-specified path {!r}".format(
                    use_flux
                )
            )
        return path
    # path = "/g/g12/corbett8/local/bin/flux"
    path = "/usr/global/tools/flux/{}/flux-c0.28.0.pre-s0.17.0.pre/bin/flux".format(
        os.getenv("SYS_TYPE")
    )
    if not os.path.isfile(path):
        path = utils.which("flux")
    if path is None:
        raise RuntimeError("Cannot find Flux installation")
    return path


def backend_dir(setup_dir, alloc_id):
    """Return the execution directory for a given alloc id."""
    return os.path.join(setup_dir, "execution_" + str(alloc_id))


class ProcessGroup(object):
    """Represents a group of `subprocess.Popen` objects."""

    def __init__(self, process_iterable):
        self._processes = process_iterable

    def poll(self):
        """Return None if one or more processes are still running."""
        if any(process.poll() is None for process in self._processes):
            return None
        return True

    def terminate(self, wait=0):
        """Kill all the processes."""
        for process in self._processes:
            try:
                process.terminate()
            except OSError:  # process has already exited
                pass
        sleep_sec = 0.1
        while wait > 0:
            if self.poll() is not None:
                break
            time.sleep(sleep_sec)
            wait -= sleep_sec


class ResourceManager(object):
    """Encapsulates high-level, general information about a resource manager.

    Subclass should expose the following attributes:
    :param identifier: string identifying the resource manager
    :param allocator: the allocators.Allocator class to use for getting an allocation
    :param batch_script_occupies_core: whether, when running a batch script, an extra
        core should be allocated for the batch script itself.
    """

    def __init__(self, _):
        pass

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}()".format(type(self).__name__)

    @classmethod
    def commands_to_launch_backend(  # pylint: disable=too-many-arguments
        cls,
        timeout,
        parallelism,
        alloc_id,
        early_stop,
        multiple,
        setup_dir,
        max_concurrency,
    ):
        """Return the shell commands to start the ensemble's backend."""

        return (
            "{py} {backend_path} -t{runtime} -a{aid} -p{parallel} "
            "-e{early_stop} --setup-dir {setup_dir} -c {max_concur} {multiple_flag}"
            ">> {backend_dir}{sep}themis_backend.log 2>&1"
        ).format(
            py=sys.executable,
            backend_path=_BACKEND_PATH,
            parallel=parallelism,
            aid=alloc_id,
            runtime=timeout,
            early_stop=early_stop,
            setup_dir=setup_dir,
            backend_dir=backend_dir(setup_dir, alloc_id),
            sep=os.sep,
            max_concur=max_concurrency,
            multiple_flag="--multiple " if multiple else "",
        )

    @classmethod
    def launch_workers(cls, parallelism, server_args, setup_dir, max_concurrency):
        """Launch workers that will submit and monitor runs.

        :param parallelism: the number of workers to launch
        :param server_args: the network connection information for the workers
        :returns: an object with `poll()` and `kill(wait)` attributes representing
            the workers.
        """
        if parallelism <= 0:
            return ProcessGroup(
                [
                    cls._launch_single_worker(
                        NoResourceManager.build_cmd(utils.Step(sys.executable)),
                        0,
                        server_args,
                        setup_dir,
                        max_concurrency,
                    )
                ]
            )
        return ProcessGroup(
            [
                cls._launch_single_worker(
                    cls.build_cmd(utils.Step(sys.executable)),
                    i,
                    server_args,
                    setup_dir,
                    math.ceil(max_concurrency / parallelism),
                )
                for i in range(parallelism)
            ]
        )

    # pylint: disable=too-many-arguments
    @classmethod
    def _launch_single_worker(
        cls, launch_command, identifier, server_args, setup_dir, max_concurrency,
    ):
        launch_command.extend(
            (
                "{py} {worker_path} --id {identifier} --server-args "
                "{server_args} --setup-dir {setup_dir} -c {max_concur}"
            )
            .format(
                py=sys.executable,
                identifier=identifier,
                worker_path=_WORKER_PATH,
                server_args=" ".join(server_args),
                setup_dir=setup_dir,
                max_concur=int(max_concurrency),
            )
            .split()
        )
        with open("themis_worker_{}.log".format(identifier), "a") as file_handle:
            return subprocess.Popen(
                launch_command, stdout=file_handle, stderr=subprocess.STDOUT,
            )

    @classmethod
    def executor(cls, setup_dir, server_args):
        """Return an Executor instance for use with this resource manager."""
        return executors.ProcessExecutor(setup_dir, server_args, cls.build_cmd)

    @classmethod
    def build_cmd(cls, step):
        """Return a command to launch an application with resource requirements."""
        raise NotImplementedError("Using abstract class")


@_register_rmgr
class Moab(ResourceManager):
    """The Moab resource manager.

    Only partially implemented because only used by the ``laf.BatchSubmitter`` class.
    """

    identifier = "moab"
    allocator = allocators.MoabAllocator

    @classmethod
    def build_cmd(cls, step):
        raise NotImplementedError("MOAB only partially implemented")


@_register_rmgr
class Sbatch(ResourceManager):
    """The Sbatch resource manager.

    Only partially implemented because only used by the ``laf.BatchSubmitter`` class.
    """

    identifier = "sbatch"
    allocator = allocators.SbatchAllocator

    @classmethod
    def build_cmd(cls, step):
        raise NotImplementedError("SBATCH only partially implemented")


@_register_rmgr
class Slurm(ResourceManager):
    """The Slurm resource manager, as it works on LC."""

    identifier = "slurm"
    allocator = allocators.SbatchAllocator
    batch_script_occupies_core = True
    _BASE_CMD = ("srun", "--exclusive", "--mpibind=off")

    @classmethod
    def build_cmd(cls, step):
        """Return the run command ('srun') for TOSS systems"""
        base_cmd = list(cls._BASE_CMD)
        if step.batch_script:
            base_cmd.extend(("-n1", "-c1", "-N1"))
        else:
            if step.tasks > 0:
                base_cmd.extend(("-n", str(step.tasks)))
            if step.cores_per_task > 0:
                base_cmd.extend(("-c", str(step.cores_per_task)))
        if step.timeout > 0:
            base_cmd.extend(("-t", str(step.timeout)))
        return base_cmd


@_register_rmgr
class LANLSlurm(Slurm):
    """LANL's Slurm is different than LC's---primarily in ``srun`` options."""

    identifier = "lanl-slurm"
    _BASE_CMD = ("srun", "--exclusive", "--gres=craynetwork:0")

    @classmethod
    def build_cmd(cls, step):
        """Return the run command ('srun') for TOSS systems"""
        if step.batch_script:
            return []
        return super(LANLSlurm, cls).build_cmd(step)


@_register_rmgr
class Flux(ResourceManager):
    """The Flux resource manager.

    Flux is an unusual resource manager because it can run on top of another
    resource manager---e.g. Slurm.
    """

    identifier = "flux"
    batch_script_occupies_core = False

    def __init__(self, path):
        super(Flux, self).__init__(None)
        # check whether this is a Flux-native machine or if another RM is running
        if default_resource_manager_id() == self.identifier:
            self.underlying_resource_mgr = None  # flux-native
        else:
            #  another RM is running
            self.underlying_resource_mgr = identify_resource_manager()
        self.path = path

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}(path={}, underlying_resource_mgr={})".format(
            type(self).__name__, self.path, self.underlying_resource_mgr
        )

    @property
    def allocator(self):
        """Return the allocator of the underlying rmgr."""
        if self.underlying_resource_mgr is not None:
            return self.underlying_resource_mgr.allocator
        return allocators.FluxAllocator

    def commands_to_launch_backend(  # pylint: disable=too-many-arguments
        self,
        timeout,
        parallelism,
        alloc_id,
        early_stop,
        multiple,
        setup_dir,
        max_concurrency,
    ):
        """Return the shell commands to start the ensemble's backend.

        Use the `flux tree` utility if needed.
        """
        base_cmd = ""
        if self.underlying_resource_mgr is not None:
            if isinstance(self.underlying_resource_mgr, Slurm):
                base_cmd = "srun --ntasks-per-node=1 --mpibind=off "
            elif isinstance(self.underlying_resource_mgr, Lsf):
                base_cmd = "lrun -T1 --mpibind=off "
            base_cmd += self.path + " start "
        return base_cmd + super(Flux, self).commands_to_launch_backend(
            timeout,
            parallelism,
            alloc_id,
            early_stop,
            multiple,
            setup_dir,
            max_concurrency,
        )

    def launch_workers(self, parallelism, server_args, setup_dir, max_concurrency):
        if parallelism == 0:
            return super(Flux, self).launch_workers(
                parallelism, server_args, setup_dir, max_concurrency
            )
        return ProcessGroup(
            [
                subprocess.Popen(
                    (
                        "flux tree -T{parallelism} -c{cores_per_node} -J{parallelism} "
                        "-- {py} {worker_path} --server-args {server_args} "
                        "-c {max_concur} --setup_dir {setup_dir}"
                    )
                    .format(
                        parallelism=parallelism,
                        cores_per_node=os.getenv("SLURM_CPUS_ON_NODE"),
                        py=sys.executable,
                        worker_path=_WORKER_PATH,
                        server_args=" ".join(server_args),
                        setup_dir=setup_dir,
                        max_concur=math.ceil(max_concurrency / parallelism),
                    )
                    .split()
                )
            ]
        )

    @classmethod
    def _stringify_topology(cls, topology_as_iterable):
        """Convert a topology to a string recognized by Flux-tree."""
        topology = [str(entry) for entry in topology_as_iterable]
        return "x".join(topology)

    @classmethod
    def _calculate_topology(cls, target_leaf_nodes, max_split=50, default_split=5):
        """Recursively calculate a topology given a target number of leaf nodes."""
        if target_leaf_nodes > max_split:
            result = [default_split]
            result.extend(
                cls._calculate_topology(math.ceil(target_leaf_nodes / default_split))
            )
            return result
        return [target_leaf_nodes]

    @classmethod
    def available(cls):
        """Return True if Flux is installed and available."""
        try:
            validate_flux_path(None)
        except ValueError:
            return False
        except RuntimeError:
            return False
        else:
            return True

    @classmethod
    def build_cmd(cls, step):
        """Return the run command for flux."""
        if step.batch_script:
            base_cmd = ["flux", "mini", "alloc"]
        else:
            base_cmd = ["flux", "mini", "run"]
        if step.tasks > 0:
            base_cmd.extend(("-n", str(step.tasks)))
        if step.cores_per_task > 0:
            base_cmd.extend(("-c", str(step.cores_per_task)))
        if step.gpus_per_task > 0:
            base_cmd.extend(("-g", str(step.gpus_per_task)))
        if step.timeout > 0:
            base_cmd.extend(("-t", str(step.timeout) + "m"))
        return base_cmd

    @classmethod
    def executor(cls, setup_dir, server_args):
        if executors.FLUX_BINDINGS_AVAIL:
            return executors.FluxBindingsExecutor(setup_dir, server_args)
        return super(Flux, cls).executor(setup_dir, server_args)


@_register_rmgr
class Lsf(ResourceManager):
    """The LSF resource manager, for Sierra and Lassen."""

    identifier = "lsf"
    allocator = allocators.BsubAllocator
    batch_script_occupies_core = True

    @classmethod
    def build_cmd(cls, step):
        """Return the run command ('lrun') for LSF systems"""
        base_cmd = ["lrun", "--pack"]
        if step.tasks > 0:
            base_cmd.extend(("-n", str(step.tasks)))
        if step.cores_per_task > 0:
            base_cmd.extend(("-c", str(step.cores_per_task)))
        if step.gpus_per_task > 0:
            base_cmd.extend(("-g", str(step.gpus_per_task)))
        return base_cmd


@_register_rmgr
class NoResourceManager(ResourceManager):
    """For systems without a resource manager running on top of the OS."""

    identifier = "none"
    allocator = allocators.InteractiveAllocator
    batch_script_occupies_core = True

    @classmethod
    def build_cmd(cls, step):
        """Return a prefix command to launch an application with a given resource set.

        Should be overridden in subclasses. Base implementation launches
        applications directly, with no regard to their resource requirements.
        """
        return []



"""Classes and functions for HTTP networking across Themis' backend.

Includes base classes, and derived classes for networking between
the backend and workers.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import ssl
import collections
import traceback
import platform
import subprocess as sp
import contextlib
import threading
import socket
import logging

from themis.database.utils import EnsembleDatabase
from themis import backend
from themis.versions import pickle, http_server, http_client, urlencode, parse_qs


CERT_NAME = "ssl_certificate.pem"
_DEFAULT_TIMEOUT = socket.getdefaulttimeout()


def create_ssl_certificate(cert_dir=os.curdir, cert_name=CERT_NAME):
    """Create an SSL certificate"""
    cert_path = os.path.join(cert_dir, cert_name)
    if not os.path.exists(cert_path):
        command = (
            (
                "openssl req -new -x509 -nodes -out {cert} -keyout {cert} "
                "-subj /C=US/ST=CA/L=Livermore/O=LLNL/OU=LLNL/CN=llnl.gov"
            )
            .format(cert=cert_path)
            .split()
        )
        subproc = sp.Popen(command, stdout=sp.PIPE, stderr=sp.STDOUT)
        stdout, _ = subproc.communicate()
        if subproc.returncode != 0:
            raise OSError(
                "SSL certificate couldn't be generated; gave error:\n{}".format(stdout)
            )
    return os.path.abspath(cert_path)


def copy_and_empty(deq):
    """Empty a collections.deque into a list and return the list."""
    while deq:
        yield deq.popleft()


def copy_and_empty2(deq):
    """Empty a deque of 2-tuples into two lists and return them."""
    copied1 = []
    copied2 = []
    while deq:
        first, second = deq.popleft()
        copied1.append(first)
        copied2.append(second)
    return (copied1, copied2)


class BaseRequestHandler(http_server.BaseHTTPRequestHandler, object):
    """Handles requests to the server. Each request instantiates an instance."""

    def log_message(self, format_str, *args):  # pylint: disable=arguments-differ
        """Log a message."""
        logging.getLogger(__name__).info(format_str, *args)

    def get_content(self):
        """Get the content from a POST request.

        Raises ValueError if Content-Length header is improper
        """
        if "Content-Length" not in self.headers:
            return None
        length = int(self.headers["Content-Length"])
        return self.rfile.read(length)

    def call_handler(self):
        """Call a handler method based on the `topic` value in the query string."""
        parsed_request = parse_qs(self.path)
        try:
            content = self.get_content()
        except ValueError:
            self.send_response(400, message="Invalid Content-Length header.")
            self.end_headers()
            # pylint: disable=attribute-defined-outside-init
            self.close_connection = True
            return
        handler = getattr(self, "handle_" + parsed_request.get("topic", ("",))[0], None)
        try:
            content_to_write = handler(parsed_request, content)
        except Exception as unknown_exc:  # pylint: disable=broad-except
            self.log_message("Invalid request:\n" + traceback.format_exc())
            self.send_response(400, message=str(unknown_exc))
            content_to_write = None
        else:
            self.send_response(200)
        self.end_headers()
        if content_to_write is not None:
            self.wfile.write(content_to_write)
        # pylint: disable=attribute-defined-outside-init
        self.close_connection = True

    def handle_ping(self, request, content):
        """For topic=ping, do nothing, just acknowledge the request."""

    def do_GET(self):  # pylint: disable=invalid-name
        """Handle a GET request."""
        self.call_handler()

    def do_POST(self):  # pylint: disable=invalid-name
        """Handle a POST request."""
        self.call_handler()


class BaseServer(object):
    """Base class for HTTP network servers.

    On construction, starts the server in a separate thread.
    """

    def __init__(self, request_handler, cert_path):
        if platform.system().lower() in ["darwin", "windows"]:
            address = "127.0.0.1"  # LLNL-issued laptops have trouble running servers
        else:
            address = socket.gethostname()
        self.server = http_server.HTTPServer((address, 0), request_handler)
        if os.path.isfile(cert_path):
            context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
            context.load_cert_chain(cert_path)
            self.server.socket = context.wrap_socket(
                self.server.socket, server_side=True
            )
        self.profiler = backend.get_profiler()
        self.server_thread = threading.Thread(
            target=backend.profile,
            name="server_thread",
            args=(self.profiler, self.server.serve_forever),
        )
        self.server_thread.daemon = True
        self.server_thread.start()

    def __repr__(self):
        """Return string representation of constructor call"""
        return "{}()".format(type(self).__name__)

    @property
    def server_address(self):
        """Return an iterable of information about the server.

        This information should be sufficient for a worker to connect.
        """
        return tuple(str(entry) for entry in self.server.server_address)

    def shutdown(self, ensemble_db):
        """Shut down the server; stop sending and receiving information.

        This should only be called once there are no more workers/workers active.
        """
        self.server.shutdown()
        self.server.server_close()
        self.server_thread.join()
        self.save_progress(ensemble_db)
        return self.profiler

    def save_progress(self, ensemble_db):
        """Save any temporary state to `ensemble_db`."""

    def is_alive(self):
        """Return True if the server is still alive and working, False otherwise."""
        return self.server_thread.is_alive()


class BaseClient(EnsembleDatabase):
    """Base class for client connections to `BaseServer` servers."""

    def __init__(self, cert_path, host, port):
        """Create a new WorkerClient instance connecting to a server at (host, port)."""
        self.host = host
        self.port = int(port)
        if os.path.isfile(cert_path):
            self.context = ssl.create_default_context(
                purpose=ssl.Purpose.SERVER_AUTH, cafile=cert_path
            )
            self.context.check_hostname = False
        else:
            self.context = None

    def __repr__(self):
        """Return string representation of constructor call"""
        return "{}({!r}, {!r}, {!r})".format(
            type(self).__name__, self.context, self.host, self.port
        )

    def ping(self, timeout=_DEFAULT_TIMEOUT):
        """Ping the server at the given host and port."""
        return self._get_response("GET", "ping", timeout=timeout)

    def _make_connection(self, timeout=_DEFAULT_TIMEOUT):
        """Make a HTTP connection to the server."""
        if self.context is not None:
            return http_client.HTTPSConnection(
                self.host, self.port, context=self.context, timeout=timeout
            )
        return http_client.HTTPConnection(self.host, self.port, timeout=timeout)

    # pylint: disable=too-many-arguments
    def _get_response(
        self, method, topic, query=None, body=None, timeout=_DEFAULT_TIMEOUT
    ):
        """Get the content of the server's response to a given query.

        :raise IOError: if the connection is unsuccessful.
        """
        query = {} if query is None else query
        query["topic"] = topic
        with contextlib.closing(self._make_connection(timeout)) as conn:
            # pylint: disable=no-member
            conn.request(method, urlencode(query), body=body)
            response = conn.getresponse()
            content = response.read()
        if response.status != 200:
            raise IOError(
                "Connection unsuccessful; error code: {}, reason: {}".format(
                    response.status, response.reason
                )
            )
        return content

    def _get_pickled_data(self, topic, query=None):
        """Unpickle and return data fetched from the server."""
        return pickle.loads(self._get_response("GET", topic, query))

    def _send_data_pickle(self, topic, query=None, data=None):
        """Pickle data and send it to the server."""
        return self._get_response(
            "POST", topic, query, pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL),
        )


class BackendRequestHandler(BaseRequestHandler):
    """Handles requests to the server. Each request instantiates an instance."""

    def handle_enable_runs(self, request, content):  # pylint: disable=unused-argument
        """Add enabled run_ids to some in-memory copy."""
        self.server.enabled_runs.extend(pickle.loads(content))

    def handle_runs_to_kill(self, request, content):  # pylint: disable=unused-argument
        """Return a pickled copy of the in-memory copy of runs to kill."""
        return pickle.dumps(self.server.runs_to_kill, protocol=pickle.HIGHEST_PROTOCOL)

    def handle_add_result(self, request, content):  # pylint: disable=unused-argument
        """Add status to some in-memory copy and wait for it to be written."""
        run_ids, results = pickle.loads(content)
        self.server.run_results.extend(zip(run_ids, results))

    def handle_set_run_status(
        self, request, content
    ):  # pylint: disable=unused-argument
        """Add status to some in-memory copy and wait for it to be written."""
        run_ids, statuses = pickle.loads(content)
        self.server.run_statuses.extend(zip(run_ids, statuses))

    def handle_new_runs(self, request, content):  # pylint: disable=unused-argument
        """Handle a worker's args POST request"""
        num_requested_new_runs = pickle.loads(content)
        return pickle.dumps(
            self.server.run_feeder.new_runs(num_requested_new_runs),
            protocol=pickle.HIGHEST_PROTOCOL,
        )


class BackendServer(BaseServer):
    """Server for communicating between the backend process and worker processes.

    Provides the subset of the themis.database.EnsembleDatabase functionality that
    workers need.
    """

    def __init__(self, cert_path, run_feeder):
        super(BackendServer, self).__init__(BackendRequestHandler, cert_path)
        self.enabled_runs = collections.deque()
        self.runs_to_kill = collections.deque()
        self.run_statuses = collections.deque()
        self.run_results = collections.deque()
        self._export_attributes_to_server(run_feeder)

    def __repr__(self):
        """Return string representation of constructor call"""
        return "{}()".format(type(self).__name__)

    def _export_attributes_to_server(self, run_feeder):
        self.server.run_feeder = run_feeder
        self.server.enabled_runs = self.enabled_runs
        self.server.runs_to_kill = self.runs_to_kill
        self.server.run_results = self.run_results
        self.server.run_statuses = self.run_statuses

    def save_progress(self, ensemble_db):
        """Write the server's caches to the database."""
        ensemble_db.set_run_status(*copy_and_empty2(self.run_statuses))
        ensemble_db.enable_runs(copy_and_empty(self.enabled_runs))
        ensemble_db.add_result(*copy_and_empty2(self.run_results))
        self.runs_to_kill.clear()
        self.runs_to_kill.extend(ensemble_db.runs_to_kill())

    def is_alive(self):
        """Return True if the server is still alive and working, False otherwise."""
        return self.server_thread.is_alive()


class BackendClient(BaseClient):
    """Represents a connection to a BackendServer instance."""

    def runs_to_kill(self):
        """Mirrors EnsembleDatabase method."""
        return self._get_pickled_data("runs_to_kill")

    def new_runs(self, limit):
        """Mirrors EnsembleDatabase method."""
        return pickle.loads(self._send_data_pickle("new_runs", data=limit))

    def enable_runs(self, run_ids):
        """Mirrors EnsembleDatabase method."""
        self._send_data_pickle("enable_runs", data=run_ids)

    def add_result(self, run_ids, results):
        """Mirrors EnsembleDatabase method."""
        self._send_data_pickle("add_result", data=(run_ids, results))

    def set_run_status(self, run_ids, statuses):
        """Mirrors EnsembleDatabase method."""
        self._send_data_pickle("set_run_status", data=(run_ids, statuses))



"""Common functions and utility functions for `themis.database`."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sqlite3 as sql
import pickle
import contextlib


def unpickle_or_none(value):
    """Unpickle a non-None value."""
    return pickle.loads(value) if value is not None else None


class EnsembleDatabase(object):  # pylint: disable=too-few-public-methods
    """Base class for all ensemble backends."""

    # A sort of enum for run completion statuses
    RUN_QUEUED = 0
    RUN_SUCCESS = 1
    RUN_FAILURE = 2
    RUN_KILLED = 3
    RUN_ABORTED = 4


@contextlib.contextmanager
def cursorcloser(conn):
    """A context manager for closing cursor objects.

    :param conn: a Python DB API 2.0 Connection object.
    """
    # Code to acquire resource, e.g.:
    cursor = conn.cursor()
    try:
        yield cursor
    finally:
        conn.commit()
        cursor.close()


@contextlib.contextmanager
def sql_transaction(cursor, immediate=False):
    """A context manager for managing SQL transactions.

    Issues a "BEGIN" statement when entered, and then
    a "COMMIT" if no errors were raised within the managed block,
    or a "ROLLBACK" otherwise.

    :param cursor: a Python DB API 2.0 Cursor object for a SQL database.
    :param immediate: used in sqlite to begin a write transaction
        immediately, not waiting for the first write statement.
    """
    if immediate:
        cursor.execute("BEGIN IMMEDIATE")
    else:
        cursor.execute("BEGIN")
    try:
        yield
    except:
        cursor.execute("ROLLBACK")
        raise
    else:
        cursor.execute("COMMIT")


class SQLiteMixin(object):
    """Defines a handful of useful methods for objects encapsulating SQLite DBs.

    A connection to the DB is opened on construction and closed
    on garbage collection.
    """

    def __init__(self, db_path, timeout=60):
        """Constructor. Opens a connection to the database.

        The connection is only closed on object destruction.
        """
        self._db_path = db_path
        self._db_handle = sql.connect(
            db_path, timeout=timeout, isolation_level=None, check_same_thread=False
        )
        self._db_handle.text_factory = str  # needed only for Python 2
        self._table_names = []

    def __del__(self):
        """Destructor. Closes the connection to the database."""
        if hasattr(self, "_db_handle"):
            # this check is necessary if the constructor raises an exception
            self._db_handle.close()

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}({!r})".format(type(self).__name__, self._db_path)

    def delete(self):
        """Empty the database of all data. Deletes all tables."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            for table in self._table_names:
                cursor.execute("DROP TABLE IF EXISTS {}".format(table))

    def _execute_shorthand(self, *query_param_pairs):
        """Helper method for executing a series of queries."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            for query, params in query_param_pairs:
                if params is not None:
                    cursor.execute(query, params)
                else:
                    cursor.execute(query)
            return cursor.fetchone()



"""This script wraps up a finished application run.

Most of what this script does
revolves around calling the user's post_run method and responding to the result.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import logging

if __name__ == "__main__":
    # when running as a script, sys.path needs to be set for imports
    sys.path.insert(0, os.path.abspath(__file__).rsplit(os.sep, 4)[0])

# pylint: disable=wrong-import-position
from themis.utils import DirectoryManager
from themis.backend.worker import utils
from themis import versions

_FAILURE = 1e99


def user_post_run(post_run, run_dir):
    """Call the user's post_run function, if it exists."""
    if callable(post_run):
        with DirectoryManager(run_dir):
            return post_run()
    return None


def main(run_id, server_args, setup_dir):
    """Invoke the user's post_run application interface function.

    Store the result in the database indirectly by passing it to a
    Themis server.
    """
    run_db, app_interface, run_dir, _, _ = utils.setup_script(
        run_id, server_args, setup_dir
    )
    logger = logging.getLogger(__name__)
    post_run = getattr(app_interface, "post_run", None)
    logger.info("Calling user's post_run")
    post_run_result = user_post_run(post_run, run_dir)  # don't shield exceptions
    if post_run_result == _FAILURE:
        raise ValueError("Received {} from post_run, exiting...".format(_FAILURE))
    if post_run_result is not None:
        logger.info(
            "Received from post_run: %s", versions.reprlib.repr(post_run_result)
        )
        run_db.add_result(run_id, post_run_result)
    logger.info("Result passed off successfully, exiting normally...")


if __name__ == "__main__":
    utils.profile_worker(main, "finisher")



"""
The :code:`ensemble.user_utils` module exports certain functions that
are designed to be called from an application interface.

Since the application interface functions (``prep_run``, etc.) take no arguments,
they do not automatically have access to any information about the ensemble.
For this reason,
certain information is exported, so that, for instance, ``prep_run`` can find out
what run it is, what the sample point is for that run, and how many total runs
there are. This information is exported to the :code:`ensemble.user_utils` module;
to access it, just add :code:`from themis import user_utils` to the top
of your application interface.

If any of these functions are called outside an application interface,
their behavior is undefined.

Many of the exported functions are designed to be used by specific
application interface functions, and their behavior may change depending
on the caller (e.g. a function only meant to be used by ``prep_ensemble``
might raise an error when called by ``post_run``);
consult the documentation of each function for more information.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import themis

_RUN_ID = None  # run ID of current run
_STATUS_DB = None  # status DB handle
_SETUP_DIR = None  # setup dir for ensemble
_RUN = None  # Run object for current run


def run_id():
    """Return the integer ID of the current run.

    For use by ``prep_run`` or ``post_run``. Returns ``None`` if called
    by ``prep_ensemble`` or ``post_ensemble``.
    """
    return _RUN_ID


def run():
    """Return the ``themis.CompositeRun`` object for the current run.

    This can be used to obtain the sample, arguments, and resource requirements
    for the current run.

    For use by ``prep_run`` or ``post_run``. Returns ``None`` if called
    by ``prep_ensemble`` or ``post_ensemble``.
    """
    return _RUN


def themis_handle():
    """Return a ``themis.Themis`` for this ensemble."""
    if _SETUP_DIR is not None:
        return themis.Themis(setup_dir=_SETUP_DIR)
    return None


def stop_ensemble():
    """Stop the ensemble from initiating any new runs.

    Existing runs will complete, and then the ensemble will exit
    after calling the user's ``post_ensemble`` (if it exists).
    """
    if _RUN_ID is not None and _STATUS_DB is not None:
        _STATUS_DB.set(stop=True)



"""This module contains the SQLiteStatusStore class."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sqlite3

from themis.database import utils
from themis.database.utils import cursorcloser, sql_transaction
from themis.resource.allocators import Allocation


class SQLiteStatusStore(utils.SQLiteMixin):
    """SQLite database holding global information about an ensemble."""

    _SELECT_ALLOC = (
        "SELECT nodes, partition, bank, name, timeout, "
        "repeats FROM Allocations WHERE id = ?"
    )

    def __init__(self, db_path):
        """Initialize a new instance."""
        super(SQLiteStatusStore, self).__init__(db_path)
        self._table_names.extend(("EnsembleStatus", "Allocations"))
        self._db_handle.row_factory = sqlite3.Row

    def create(self):
        """Create the EnsembleStatus table.

        This table contains all the ensemble status information, which is described in
        the `get_ensemble_status` method.
        """
        self._execute_shorthand(
            (
                "CREATE TABLE EnsembleStatus (id SMALLINT PRIMARY KEY, "
                "prepped SMALLINT, stop SMALLINT, exit SMALLINT, "
                "failed_runs INTEGER)",
                None,
            ),
            ("INSERT INTO EnsembleStatus VALUES (0, 0, 0, 0, 0)", None),
            (
                "CREATE TABLE Allocations (id INTEGER PRIMARY KEY, nodes INTEGER, "
                "partition TEXT, bank TEXT, name TEXT, timeout REAL, "
                "repeats INTEGER)",
                None,
            ),
        )

    def fetch(self):
        """Return a mapping containing all the ensemble status information.

        This information consists of the following value;
        * A boolean indicating whether prep_ensemble has been called
        * A boolean indicating whether the ensemble should be stopped
        * A boolean indicating whether the ensemble may exit
        * An integer number of ensemble-wide failed runs
        * An integer number of allocation repeats remaining for the ensemble
        """
        return self._execute_shorthand(
            ("SELECT * FROM EnsembleStatus WHERE id = 0", None)
        )

    def set(self, **kwargs):
        """Update the ensemble status, setting `field = value`."""
        with utils.cursorcloser(self._db_handle) as cursor:
            with utils.sql_transaction(cursor):
                for key, value in kwargs.items():
                    cursor.execute(
                        "UPDATE EnsembleStatus SET {key} = (?) WHERE id = 0".format(
                            key=key
                        ),
                        (value,),
                    )

    def increment_failed_runs(self, increment):
        """Increment the number of ensemble-wide failed runs by `increment`.

        Return the new value.
        """
        return self._execute_shorthand(
            (
                "UPDATE EnsembleStatus SET failed_runs = failed_runs + ? WHERE id = 0",
                (increment,),
            ),
            ("SELECT failed_runs FROM EnsembleStatus WHERE id = 0", None),
        )[0]

    def add_allocation(self, alloc):
        """Add an allocation to the database and return its ID."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.execute(
                "INSERT INTO Allocations(nodes, partition, bank, name, timeout, "
                "repeats) VALUES (?, ?, ?, ?, ?, ?)",
                (
                    alloc.nodes,
                    alloc.partition,
                    alloc.bank,
                    alloc.name,
                    alloc.timeout,
                    alloc.repeats,
                ),
            )
            return cursor.lastrowid

    def get_allocation(self, alloc_id):
        """Get the allocation specified by the id."""
        row = self._execute_shorthand((self._SELECT_ALLOC, (alloc_id,)))
        return Allocation(**row)

    def decrement_repeats(self, alloc_id):
        """Decrement the number of allocation repeats left for an allocation.

        Return the new value.
        """
        return Allocation(
            **self._execute_shorthand(
                (
                    "UPDATE Allocations SET repeats = repeats - 1 WHERE id = ?",
                    (alloc_id,),
                ),
                (self._SELECT_ALLOC, (alloc_id,)),
            )
        )
